package com.ssi.model;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.action.ActionFuture;
import org.elasticsearch.action.admin.indices.cache.clear.ClearIndicesCacheRequest;
import org.elasticsearch.action.admin.indices.cache.clear.ClearIndicesCacheResponse;
import org.elasticsearch.action.admin.indices.stats.IndexStats;
import org.elasticsearch.action.admin.indices.stats.IndicesStatsRequest;
import org.elasticsearch.action.admin.indices.stats.IndicesStatsResponse;
import org.elasticsearch.action.search.SearchRequestBuilder;
import org.elasticsearch.client.transport.TransportClient;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.reindex.DeleteByQueryAction;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.joda.time.Days;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.*;

/**
 * Description:
 *
 * @author LiXiaoCong
 * @version 2019/1/17 15:07
 */
@Component
@Slf4j
public class ElasticSearchMetaModel {

  private final static DateTimeZone TIME_ZONE = DateTimeZone.forID("Asia/Shanghai");

  @Autowired
  private TransportClient transportClient;

  public Map<String, IndexStats> getIndicesStats() {
    ActionFuture<IndicesStatsResponse> isr = transportClient.admin().indices()
        .stats(new IndicesStatsRequest().all());
    return isr.actionGet().getIndices();
  }

  public List<String> getIndices() {
    return Lists.newArrayList(getIndicesStats().keySet());
  }

  public boolean exist(String index) {
    List<String> indexs = getIndices();
    return indexs == null ? false : indexs.contains(index);
  }

  public String[] selectExist(String... indexs) {
    return selectExist(Sets.newHashSet(indexs));
  }

  public String[] selectExist(Collection<String> indexs) {
    List<String> all = getIndices();
    if (all == null) {
      all = Lists.newArrayList();
    }
    indexs.retainAll(all);
    return indexs.toArray(new String[indexs.size()]);
  }


  public SearchRequestBuilder newSearchRequestBuilderInDayIndex(Long startTime,
      Long stopTime,
      String indexBase,
      int defaultDayPeriod,
      boolean defaultContainsToday) {

    return newSearchRequestBuilderInDayIndex(startTime, stopTime, indexBase,
        defaultDayPeriod, defaultContainsToday, "day");
  }

  public SearchRequestBuilder newSearchRequestBuilderInDayIndex(Long startTime,
      Long stopTime,
      String indexBase,
      int defaultDayPeriod,
      boolean defaultContainsToday,
      String indexRule) {
    SearchRequestBuilder requestBuilder = null;

    if (indexRule == null) {
      requestBuilder = transportClient.prepareSearch(indexBase);
    } else if (startTime == null && stopTime == null) {
      requestBuilder = transportClient.prepareSearch(String.format("%s_2*", indexBase));
    } else {
      String[] indexs = getIndexs(startTime, stopTime, indexBase, defaultDayPeriod,
          defaultContainsToday, indexRule);
      if (indexs.length > 0) {
        log.info(String.format("查询es的索引有：%s", Lists.newArrayList(indexs)));
        requestBuilder = transportClient.prepareSearch(indexs);
      }
    }
    return requestBuilder;
  }


  public String[] getIndexs(Long startTime,
      Long stopTime,
      String indexBase,
      int defaultDayPeriod,
      boolean defaultContainsToday) {
    return getIndexs(startTime, stopTime, indexBase, defaultDayPeriod, defaultContainsToday, "day");
  }

  public String[] getIndexs(Long startTime,
      Long stopTime,
      String indexBase,
      int defaultDayPeriod,
      boolean defaultContainsToday,
      String indexRule) {

    if (indexRule == null) {
      return new String[]{indexBase};
    }

    DateTime stopDate;
    if (stopTime == null) {
      DateTime now = DateTime.now(TIME_ZONE);
      if (defaultContainsToday) {
        stopDate = now;
      } else {
        stopDate = now.minusDays(1);
      }
    } else {
      stopDate = new DateTime(stopTime.longValue(), TIME_ZONE);
    }

    DateTime startDate = (startTime == null ? stopDate.minusDays(defaultDayPeriod)
        : new DateTime(startTime.longValue(), TIME_ZONE));

    List<String> indexList = new ArrayList<>();
    String[] indexs;
    int days = Days.daysBetween(startDate, stopDate).getDays();
    Set<String> indexSet = new HashSet<>();
    if (days >= 0) {
      for (int i = 0; i <= days; i++) {
        try {
          switch (indexRule.toLowerCase()) {
            //按年分区
            case "year":
              indexSet.add(String.format("%s_%s*",
                  indexBase, startDate.plusDays(i).toString("yyyy")));
              break;
            //按月分区
            case "month":
              indexSet.add(String.format("%s_%s*",
                  indexBase, startDate.plusDays(i).toString("yyyyMM")));
              break;
            //按天分区
            case "day":
              indexList.add(String.format("%s_%s*",
                  indexBase, startDate.plusDays(i).toString("yyyyMMdd")));
              break;
            default:
              indexSet.add(indexBase);
          }
        } catch (Exception e) {
        }
      }
      indexList.addAll(indexSet);
      indexs = indexList.toArray(new String[indexList.size()]);
    } else {
      indexs = new String[0];
    }

    log.info("查询索引有：indexs = " + Lists.newArrayList(indexs).toString());

    return indexs;
  }

  /**
   * 清空指定索引的缓存
   *
   * @param indexName 索引名
   * @return 清空结果
   */
  public boolean clearIndicesCache(String... indexName) {
    ClearIndicesCacheResponse response;
    try {
      response = transportClient.admin().indices()
          .clearCache(new ClearIndicesCacheRequest(indexName)
              //.fieldDataCache(true)
              .queryCache(true)
          ).actionGet();
      if (response.getFailedShards() > 0) {
        return false;
      }
      return true;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return false;
  }


  public String[] deleteIndexByIds(List<Map<String, Object>> deleteMaps) {

    String[] idArray = null;
    if (deleteMaps != null) {
      Set<String> indexs = Sets.newHashSet();
      Set<String> ids = Sets.newHashSet();

      deleteMaps.forEach(deleteMap -> {

        String id = String.valueOf(deleteMap.get("id"));
        String index = String.valueOf(deleteMap.get("index"));
        String table = String.valueOf(deleteMap.get("table"));

        if (id != null && index != null && table != null) {
          ids.add(id);
          indexs.add(index);
        }
      });

      String[] indices = indexs.toArray(new String[indexs.size()]);
      idArray = ids.toArray(new String[ids.size()]);
      if (indices.length > 0) {

        DeleteByQueryAction.INSTANCE.newRequestBuilder(transportClient)
            .filter(QueryBuilders.termsQuery("id", ids))
            .source(indices)
            .get();

        clearIndicesCache(indices);
      }
    }

    return idArray;
  }
}
