package com.ssi.service.impl;

import cn.afterturn.easypoi.excel.entity.ExportParams;
import cn.afterturn.easypoi.excel.export.ExcelBatchExportService;
import cn.afterturn.easypoi.handler.impl.ExcelDataHandlerDefaultImpl;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ssi.constant.enums.CommandResultEnum;
import com.ssi.constant.enums.CommandTypeEnum;
import com.ssi.entity.VmsCarControlCommand;
import com.ssi.entity.VmsVehicle;
import com.ssi.entity.dto.CarControlCommandParam;
import com.ssi.entity.vo.CarControlCommandVo;
import com.ssi.mapper.VmsCarControlCommandMapper;
import com.ssi.mapper.VmsVehicleMapper;
import com.ssi.model.RedisDataModel;
import com.ssi.response.AccessResponse;
import com.ssi.response.SSIPage;
import com.ssi.response.SSIResponse;
import com.ssi.service.VmsCarControlCommandService;
import com.ssi.utils.AuthorizationUtils;
import com.ssi.utils.PoiUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import org.springframework.util.Assert;
import org.springframework.web.client.RestTemplate;

import javax.servlet.http.HttpServletResponse;
import java.util.*;

/**
 * 车辆远程控制服务实现类
 */
@Service("vmsCarControlCommandService")
@Slf4j
public class VmsCarControlCommandServiceImpl extends ServiceImpl<VmsCarControlCommandMapper, VmsCarControlCommand> implements VmsCarControlCommandService {

    @Autowired
    private RestTemplate restTemplate;
    @Autowired
    private VmsVehicleMapper vmsVehicleMapper;
    @Autowired
    private RedisDataModel redisDataModel;

    @Value("${command-url}")
    private String commandUrl;

    /**
     * 远程控制指令分页数据
     * @param vo
     * @return
     */
    @Override
    public SSIPage queryPage(CarControlCommandVo vo) {
        IPage<VmsCarControlCommand> page = new Page<>(vo.getPageIndex(), vo.getPageSize());
        LambdaQueryWrapper<VmsCarControlCommand> wrapper = new QueryWrapper<VmsCarControlCommand>().lambda();
        wrapper.eq(vo.getCommand() != null, VmsCarControlCommand::getCommand, vo.getCommand())
                .like(vo.getVehicleNum() != null, VmsCarControlCommand::getVehicleNum, vo.getVehicleNum())
                .ge(vo.getQueryStartTime() != null, VmsCarControlCommand::getCreateTime, vo.getQueryStartTime())
                .le(vo.getQueryEndTime() != null, VmsCarControlCommand::getCreateTime, vo.getQueryEndTime())
                .orderByDesc(VmsCarControlCommand::getCreateTime);
        return new SSIPage(this.page(page, wrapper));

    }

    /**
     * 导出远程控制指令到Excel
     * @param carControlCommandVo
     * @param response
     */
    @Override
    public void export(CarControlCommandVo carControlCommandVo, HttpServletResponse response) {
        int exportType = carControlCommandVo.getExportType();
        if (exportType == 0) {
            SSIPage page = queryPage(carControlCommandVo);
            List<VmsCarControlCommand> list = (List<VmsCarControlCommand>) page.getList();
            ExcelDataHandlerDefaultImpl<Object> excelDataHandler = new ExcelDataHandlerDefaultImpl<Object>() {
                @Override
                public Object exportHandler(Object obj, String name, Object value) {
                    return super.exportHandler(obj, name, value);
                }
            };
            PoiUtils.exportExcel(excelDataHandler, list, "远程指令清单", "VmsCarControlCommand", VmsCarControlCommand.class, "远程指令清单" + ".xls", true, response);
        } else {
            ExcelBatchExportService batchExportService = new ExcelBatchExportService();
            ExportParams exportParams = new ExportParams("远程指令清单", "远程指令清单");
            exportParams.setMaxNum(1000000);
            batchExportService.init(exportParams, VmsCarControlCommand.class);
            this.getBaseMapper().getListAsStream(carControlCommandVo, resultContext -> {
                VmsCarControlCommand command = resultContext.getResultObject();
                batchExportService.appendData(Arrays.asList(command));
            });
            Workbook workbook = batchExportService.closeExportBigExcel();
            PoiUtils.downLoadExcel("远程指令清单.xlsx", response, workbook);
        }

    }

    /**
     * 执行控车指令
     * @param param 控车指令参数
     */
    @Override
    public SSIResponse executeCommand(CarControlCommandParam param) {
        VmsCarControlCommand command = new VmsCarControlCommand();
        Date date = new Date();
        String vin = param.getVin();
        VmsVehicle vehicle = vmsVehicleMapper.selectOne(new LambdaQueryWrapper<VmsVehicle>()
                .eq(VmsVehicle::getVin, vin));
        if (vehicle == null) {
            return SSIResponse.no("车辆不存在");
        }
        CommandTypeEnum commandTypeEnum = CommandTypeEnum.find(param.getCommandType());
        if (commandTypeEnum == null) {
            return SSIResponse.no("不支持的指令类型");
        }
        command.setVin(vin);
        command.setVehicleNum(vehicle.getVehicleNum());
        command.setCommand(param.getCommandType());
        command.setCreateTime(date);
        JSONObject curUser = AuthorizationUtils.getCurUser();
        Integer userId = (Integer) curUser.get("userId");
        String username = (String) curUser.get("username");
        command.setOperatorId(userId);
        command.setOperatorName(username);
        String url = commandUrl + "/command/sendF004";
        param.setCollectTime(date.getTime());
        AccessResponse accessResponse = restTemplate.postForObject(url, param, AccessResponse.class);
        if (accessResponse.isSuccess()) {
            Map data = (Map) accessResponse.getData();
            Integer commandResult = (Integer) data.get("commandResult");
            String description = CommandResultEnum.find(commandResult).getDescription();
            command.setExecuteStatus(commandResult);
            command.setExecuteMessage(description);
            this.save(command);
            if (CommandResultEnum.isExecuteSuccess(commandResult)) {
                return SSIResponse.ok();
            } else {
                return SSIResponse.no(description);
            }
        } else {
            String msg = accessResponse.getStatus().getDetails();
            command.setExecuteStatus(CommandResultEnum.FAIL.getCode());
            command.setExecuteMessage(msg);
            this.save(command);
            return SSIResponse.no(msg);
        }
    }

    /**
     * 步进
     * @param vin 车辆vin
     * @param direction 方向 1-前进 ；2-后退
     * @param distance 行驶距离(单位：CM)
     * @return
     */
    @Override
    public SSIResponse stepping(String vin, Integer direction, Integer distance) {
        Assert.notNull(vin,"车辆vin不能为空");
        Assert.notNull(direction,"方向参数direction不能为空");
        Assert.notNull(distance,"行驶距离参数distance不能为空");
        log.info(String.format("调用步进接口，vin：%s，direction：%s，distance：%s", vin, direction, distance));
        String url = commandUrl + "/remote/app/stepping";
        Map param = new HashMap();
        param.put("vin", vin);
        param.put("direction", direction);
        param.put("distance", distance);
        AccessResponse accessResponse = restTemplate.postForObject(url, param, AccessResponse.class);
        if (accessResponse.isSuccess()) {
            log.info(String.format("调用步进接口成功，vin：%s，direction：%s，distance：%s", vin, direction, distance));
            return SSIResponse.ok();
        } else {
            String msg = accessResponse.getStatus().getDetails();
            log.info(String.format("调用步进接口失败，vin：%s，direction：%s，distance：%s", vin, direction, distance));
            return SSIResponse.no(msg);
        }
    }
}
