package com.ssi.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.ssi.mapper.VmsFaultCodeMapper;
import com.ssi.response.SSIPage;
import com.ssi.response.SSIResponse;
import com.ssi.utils.PoiUtils;
import com.ssi.utils.QueryUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.springframework.stereotype.Service;

import java.io.InputStream;
import java.net.URLEncoder;
import java.util.List;
import java.util.Map;

import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;

import com.ssi.entity.VmsFaultCode;
import com.ssi.service.VmsFaultCodeService;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletResponse;

/**
 * 故障码服务实现类
 */
@Service("faultCodeService")
public class VmsFaultCodeServiceImpl extends ServiceImpl<VmsFaultCodeMapper, VmsFaultCode> implements VmsFaultCodeService {

    private static String COMMA = ",";

    private static String LINENUM = "第%s行:";

    private static String END = "; <br/>";

    /**
     * 故障码列表分页
     */
    @Override
    public SSIPage queryPage(Map<String, Object> params) {
        Integer faultGrade = (Integer) params.get("faultGrade");
        String faultCode = (String) params.get("faultCode");
        IPage<VmsFaultCode> page = this.page(
                new QueryUtils<VmsFaultCode>().getPage(params),
                new LambdaQueryWrapper<VmsFaultCode>().eq(faultGrade != null, VmsFaultCode::getFaultGrade, faultGrade)
                        .like(StringUtils.isNotBlank(faultCode), VmsFaultCode::getFaultCode, faultCode)
        );

        return new SSIPage(page);
    }

    @Override
    public void downloadExcel(InputStream in, HttpServletResponse response, String fileName) throws Exception {
        Workbook workbook = getWorkbook(in, fileName);
        response.setCharacterEncoding("UTF-8");
        response.setHeader("content-Type", "application/vnd.ms-excel");
        response.setHeader("Content-Disposition",
                "attachment;filename=" + URLEncoder.encode(fileName, "UTF-8"));
        workbook.write(response.getOutputStream());
    }

    /**
     * 导入故障码
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public SSIResponse importList(MultipartFile file) {
        List<VmsFaultCode> vmsFaultCodes = PoiUtils.importExcel(file, 0, 1, VmsFaultCode.class);
        StringBuilder msgSb = new StringBuilder();
        checkImportEmpty(vmsFaultCodes,msgSb);
        if(StringUtils.isNotBlank(msgSb.toString())){
            return SSIResponse.no(msgSb.toString());
        }
        msgSb = checkRepeat(vmsFaultCodes);
        if(StringUtils.isNotBlank(msgSb.toString())){
            return SSIResponse.no(msgSb.toString());
        }
        this.saveBatch(vmsFaultCodes);
        return SSIResponse.ok();
    }

    private void checkImportEmpty(List<VmsFaultCode> vmsFaultCodes, StringBuilder msgSb) {
        for(int i = 0; i < vmsFaultCodes.size(); i++){
            VmsFaultCode vmsFaultCode = vmsFaultCodes.get(i);
            if(vmsFaultCode != null){
                StringBuilder lineMsg = checkParamNull(vmsFaultCode);
                StringBuilder lineMsg2 = checkCodeAndDesc(vmsFaultCode);
                lineMsg.append(lineMsg2);
                if(lineMsg.length() > 0){
                    msgSb.append(String.format(LINENUM,i + 1)).append(lineMsg).append(END);
                }
            }
        }
    }

    private StringBuilder checkRepeat(List<VmsFaultCode> vmsFaultCodes) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < vmsFaultCodes.size(); i++) {
            VmsFaultCode vmsFaultCode = vmsFaultCodes.get(i);
            String fmi = vmsFaultCode.getFmi();
            String spn = vmsFaultCode.getSpn();
            String sa = vmsFaultCode.getSa();
            Integer count = this.lambdaQuery().eq(VmsFaultCode::getFmi, fmi)
                    .eq(VmsFaultCode::getSpn, spn)
                    .eq(VmsFaultCode::getSa, sa)
                    .count();
            if(count > 0){
                sb.append("FMI:").append(fmi).append(",")
                        .append("SPN:").append(spn).append(",")
                        .append("SA:").append(sa)
                        .append("在数据库中已经存在");
                return sb;
            }
            for (int j = i + 1; j < vmsFaultCodes.size(); j++) {
                VmsFaultCode code = vmsFaultCodes.get(j);
                if(fmi.equals(code.getFmi()) && spn.equals(code.getSpn()) && sa.equals(code.getSa())){
                    sb.append("FMI:").append(fmi).append(",")
                            .append("SPN:").append(spn).append(",")
                            .append("SA:").append(sa)
                            .append("在excel中有重复值");
                    return sb;
                }
            }
        }
        return sb;
    }

    private StringBuilder checkParamNull(VmsFaultCode vmsFaultCode) {
        StringBuilder lineMsg = new StringBuilder();

        if(StringUtils.isBlank(vmsFaultCode.getFaultCode())){
            lineMsg.append("故障码为空").append(COMMA);
        }if(StringUtils.isBlank(vmsFaultCode.getFaultName())){
            lineMsg.append("故障名称为空").append(COMMA);
        }if(StringUtils.isBlank(vmsFaultCode.getFaultGrade())){
            lineMsg.append("故障等级为空").append(COMMA);
        }if(StringUtils.isBlank(vmsFaultCode.getSpn())){
            lineMsg.append("SPN为空").append(COMMA);
        }if(StringUtils.isBlank(vmsFaultCode.getFmi())){
            lineMsg.append("FMI为空").append(COMMA);
        }if(StringUtils.isBlank(vmsFaultCode.getSa())){
            lineMsg.append("SA为空").append(COMMA);
        }if(StringUtils.isBlank(vmsFaultCode.getDescribition())){
            lineMsg.append("故障描述为空").append(COMMA);
        }

        if(lineMsg.length() > 0){
            lineMsg = lineMsg.deleteCharAt(lineMsg.length() - 1);
        }

        return lineMsg;
    }

    private StringBuilder checkCodeAndDesc(VmsFaultCode vmsFaultCode) {
        StringBuilder lineMsg = new StringBuilder();

        if(StringUtils.isNotBlank(vmsFaultCode.getDescribition()) && vmsFaultCode.getDescribition().length() > 200){
            lineMsg.append("故障描述大于200字").append(COMMA);
        }

        if(lineMsg.length() > 0){
            lineMsg = lineMsg.deleteCharAt(lineMsg.length() - 1);
        }
        return lineMsg;

    }

    private Workbook getWorkbook(InputStream inStr, String fileName) throws Exception {
        Workbook workbook = null;
        String fileType = fileName.substring(fileName.lastIndexOf("."));
        if (".xls".equals(fileType)) {
            workbook = new HSSFWorkbook(inStr);
        } else if (".xlsx".equals(fileType)) {
            workbook = new XSSFWorkbook(inStr);
        } else {
            throw new Exception("请上传excel文件！");
        }
        return workbook;
    }

}
