package com.ssi.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ssi.entity.VmsTosOrders;
import com.ssi.entity.VmsVehicle;
import com.ssi.entity.VmsTosOrderDailyKpi;
import com.ssi.entity.vo.VehicleKpiVo;
import com.ssi.mapper.VmsTosOrderDailyKpiMapper;
import com.ssi.service.VmsTosOrderDailyKpiService;
import com.ssi.service.VmsTosOrdersService;
import com.ssi.service.VmsVehicleService;
import com.ssi.utils.DateUtils;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @description:
 * @author: dt
 * @create: 2022-10-28 16:58
 */
@Service
public class VmsTosOrderDailyKpiServiceImpl extends ServiceImpl<VmsTosOrderDailyKpiMapper, VmsTosOrderDailyKpi> implements VmsTosOrderDailyKpiService {

    private final Logger log = LoggerFactory.getLogger(VmsTosOrderDailyKpiServiceImpl.class);

    private VmsTosOrdersService vmsTosOrdersService;

    private VmsVehicleService vmsVehicleService;

    private StringRedisTemplate redisTemplate;

    /**
     * redis key
     */
    @Value("${harbor.order.dailyKpi:harbor:order:dailyKpi}")
    private String redisDailyKpiKey;

    public VmsTosOrderDailyKpiServiceImpl(VmsTosOrdersService vmsTosOrdersService, VmsVehicleService vmsVehicleService, StringRedisTemplate redisTemplate) {
        this.vmsTosOrdersService = vmsTosOrdersService;
        this.vmsVehicleService = vmsVehicleService;
        this.redisTemplate = redisTemplate;
    }

    /**
     * 预处理任务
     * @param vins
     * @return
     */
    @Override
    public Boolean analysisTosOrderDailyKpi(String vins) {

        LambdaQueryWrapper<VmsVehicle> vmsVehicleLambdaQueryWrapper = new LambdaQueryWrapper<>();
        if (StringUtils.isNotEmpty(vins)) {
            vmsVehicleLambdaQueryWrapper.in(VmsVehicle::getVin, vins.split(","));
        }
        List<VmsVehicle> vehicles = vmsVehicleService.list(vmsVehicleLambdaQueryWrapper.eq(VmsVehicle::getStatus, 0));

        if (CollectionUtils.isEmpty(vehicles)) {
            log.info("预处理每日 KPI 任务，查询的车辆信息为空...");
            return false;
        }

        LocalDateTime now = LocalDateTime.now();
        vehicles.parallelStream().forEach(vehicle -> {
            String updateTime = redisTemplate.<String, String>boundHashOps(redisDailyKpiKey).get(vehicle.getVin());
            log.info("------ 开始处理 vin = "+ vehicle.getVin()+", updateTime = "+ updateTime +", 时间 = "+ LocalDateTime.now().toString() +" ------");
            LambdaQueryWrapper<VmsTosOrders> vmsTosOrdersLambdaQueryWrapper = new LambdaQueryWrapper<VmsTosOrders>()
                    .eq(VmsTosOrders::getVin, vehicle.getVin());
            if (StringUtils.isNotEmpty(updateTime)) {
                vmsTosOrdersLambdaQueryWrapper.ge(StringUtils.isNotEmpty(updateTime), VmsTosOrders::getUpdateTime, LocalDateTime.ofInstant(Instant.ofEpochMilli(Long.parseLong(updateTime)),
                                ZoneId.systemDefault()))
                        .lt(StringUtils.isNotEmpty(updateTime), VmsTosOrders::getUpdateTime, now);
            }
            List<VmsTosOrders> vmsTosOrders = vmsTosOrdersService.list(
                    vmsTosOrdersLambdaQueryWrapper.in(VmsTosOrders::getVehicleTaskLabel, 1 , 2, 3)
                    .in(VmsTosOrders::getTaskLocationType,1,5));

            if (CollectionUtils.isNotEmpty(vmsTosOrders)) {

                Map<String, List<VmsTosOrders>> vmsTosOrderMap = vmsTosOrders.stream().filter(vmsTosOrder -> Objects.nonNull(vmsTosOrder.getStartTime()))
                        .collect(Collectors.groupingBy(vmsTosOrder -> getDateHourKey(vmsTosOrder)));
                vmsTosOrderMap.forEach((k,v) -> {
                    VmsTosOrderDailyKpi vmsTosOrderDailyKpi = getVmsTosOrderDailyKpi(k, v, vehicle);
                    saveOrUpdateVmsTosOrderDailyKpi(vmsTosOrderDailyKpi);
                });

            } else {
                log.info("vin = "+ vehicle.getVin()+", updateTime = "+ updateTime+", 未查询到相关指令数据！");
            }

            redisTemplate.boundHashOps(redisDailyKpiKey).put(vehicle.getVin(), String.valueOf(now.atZone(ZoneId.systemDefault()).toInstant().toEpochMilli()));
            log.info("------ 结束处理 vin = "+ vehicle.getVin() +", 时间 = "+ LocalDateTime.now().toString() +" ------");
        });

        return true;
    }

    /**
     * 得到分组的 key
     * @param vmsTosOrders
     * @return
     */
    private String getDateHourKey(VmsTosOrders vmsTosOrders) {
        return DateUtils.format(vmsTosOrders.getStartTime(), "yyyy-MM-dd HH");
    }

    /**
     * 生成任务分析的结果类
     * @param vmsTosOrders
     * @return
     */
    private VmsTosOrderDailyKpi getVmsTosOrderDailyKpi(String key, List<VmsTosOrders> vmsTosOrders, VmsVehicle vehicle) {

        VmsTosOrderDailyKpi vmsTosOrderDailyKpi = new VmsTosOrderDailyKpi();
        vmsTosOrderDailyKpi.setVin(vehicle.getVin());
        vmsTosOrderDailyKpi.setVehicleNum(vehicle.getVehicleNum());
        String[] dates = key.split(" ");
        vmsTosOrderDailyKpi.setAnalysisDate(dates[0]);
        vmsTosOrderDailyKpi.setAnalysisHour(Integer.parseInt(dates[1]));
        getContainers(vmsTosOrders, vmsTosOrderDailyKpi);
        getFinishedOrderNum(vmsTosOrders, vmsTosOrderDailyKpi);
        getOrderNum(vmsTosOrders, vmsTosOrderDailyKpi);
        return vmsTosOrderDailyKpi;

    }

    /**
     * 填充 containerNum 和 containerIds
     * @param vmsTosOrders
     * @return
     */
    private void getContainers(List<VmsTosOrders> vmsTosOrders,VmsTosOrderDailyKpi vmsTosOrderDailyKpi) {
        Set<String> containerIds = vmsTosOrders.stream().filter(vmsTosOrder -> (Objects.equals(1, vmsTosOrder.getTaskLocationType()) && Objects.equals(2, vmsTosOrder.getVehicleTaskLabel())) ||
                        (Objects.equals(5, vmsTosOrder.getTaskLocationType()) && Objects.equals(1, vmsTosOrder.getVehicleTaskLabel()))).map(VmsTosOrders::getContainerId)
                .filter(StringUtils::isNotEmpty).collect(Collectors.toSet());
        vmsTosOrderDailyKpi.setContainerIds(String.join(",", containerIds));
    }

    /**
     * 填充结束任务
     * @param vmsTosOrders
     * @param vmsTosOrderDailyKpi
     */
    private void getFinishedOrderNum(List<VmsTosOrders> vmsTosOrders,VmsTosOrderDailyKpi vmsTosOrderDailyKpi) {
        String finishOrderIds = vmsTosOrders.stream().filter(vmsTosOrder -> Objects.equals(37, vmsTosOrder.getStatus()))
                .map(vmsTosOrder -> String.valueOf(vmsTosOrder.getId())).collect(Collectors.joining(","));
        vmsTosOrderDailyKpi.setFinishedOrderIds(finishOrderIds);
    }

    /**
     * 填充任务数量
     * @param vmsTosOrders
     * @param vmsTosOrderDailyKpi
     */
    private void getOrderNum(List<VmsTosOrders> vmsTosOrders,VmsTosOrderDailyKpi vmsTosOrderDailyKpi) {
        String orderIds = vmsTosOrders.stream().map(vmsTosOrder -> String.valueOf(vmsTosOrder.getId()))
                .collect(Collectors.joining(","));
        vmsTosOrderDailyKpi.setOrderIds(orderIds);
    }

    /**
     * 保存或更新
     * @param vmsTosOrderDailyKpi
     * @return
     */
    private Boolean saveOrUpdateVmsTosOrderDailyKpi(VmsTosOrderDailyKpi vmsTosOrderDailyKpi) {

        VmsTosOrderDailyKpi orderDailyKpi = this.getOne(new LambdaQueryWrapper<VmsTosOrderDailyKpi>()
                .eq(VmsTosOrderDailyKpi::getAnalysisDate, vmsTosOrderDailyKpi.getAnalysisDate())
                .eq(VmsTosOrderDailyKpi::getVin, vmsTosOrderDailyKpi.getVin())
                .eq(VmsTosOrderDailyKpi::getAnalysisHour, vmsTosOrderDailyKpi.getAnalysisHour()));

        if (Objects.nonNull(orderDailyKpi)) {
            String containerIds = Arrays.stream((orderDailyKpi.getContainerIds() + "," + vmsTosOrderDailyKpi.getContainerIds()).split(",")).distinct().filter(StringUtils::isNotEmpty).collect(Collectors.joining(","));
            vmsTosOrderDailyKpi.setContainerIds(containerIds);

            String finishedOrderIds = Arrays.stream((orderDailyKpi.getFinishedOrderIds() + "," + vmsTosOrderDailyKpi.getFinishedOrderIds()).split(",")).distinct().filter(StringUtils::isNotEmpty).collect(Collectors.joining(","));
            vmsTosOrderDailyKpi.setFinishedOrderIds(finishedOrderIds);

            String orderIds = Arrays.stream((orderDailyKpi.getOrderIds() + "," + vmsTosOrderDailyKpi.getOrderIds()).split(",")).distinct().filter(StringUtils::isNotEmpty).collect(Collectors.joining(","));
            vmsTosOrderDailyKpi.setOrderIds(orderIds);

            return this.update(vmsTosOrderDailyKpi, new LambdaUpdateWrapper<VmsTosOrderDailyKpi>()
                    .eq(VmsTosOrderDailyKpi::getAnalysisDate, vmsTosOrderDailyKpi.getAnalysisDate())
                    .eq(VmsTosOrderDailyKpi::getVin, vmsTosOrderDailyKpi.getVin())
                    .eq(VmsTosOrderDailyKpi::getAnalysisHour, vmsTosOrderDailyKpi.getAnalysisHour()));
        } else {
            return this.save(vmsTosOrderDailyKpi);
        }

    }

    /**
     * 查询分析好的 order kpi 任务
     * @param day
     * @param vehicleNum
     * @return
     */
    @Override
    public List<Map<String, String>> getOrderDailyKpi(String day, String vehicleNum) {

        List<VmsVehicle> vehicles = vmsVehicleService.list(new LambdaQueryWrapper<VmsVehicle>()
                .eq(VmsVehicle::getStatus, 0)
                .like(StringUtils.isNotBlank(vehicleNum), VmsVehicle::getVehicleNum, vehicleNum)
                .orderByAsc(VmsVehicle::getVehicleNum));

        if (CollectionUtils.isNotEmpty(vehicles)) {
            String[] vins = vehicles.stream().map(VmsVehicle::getVin).toArray(String[]::new);
            List<VmsTosOrderDailyKpi> orderDailyKpis = this.list(new LambdaQueryWrapper<VmsTosOrderDailyKpi>()
                    .eq(VmsTosOrderDailyKpi::getAnalysisDate, day)
                    .in(VmsTosOrderDailyKpi::getVin, vins));

            Map<String, List<VmsTosOrderDailyKpi>> orderDailyKpiMap = orderDailyKpis.stream().collect(Collectors.groupingBy(VmsTosOrderDailyKpi::getVin));

            List<Map<String, String>> resultList = vehicles.stream().map(vmsVehicle -> {
                try {
                    VehicleKpiVo vehicleKpiVo = new VehicleKpiVo(vmsVehicle.getVin(),vmsVehicle.getVehicleNum());
                    vehicleKpiVo.setDate(day);
                    List<VmsTosOrderDailyKpi> vmsTosOrderDailyKpis = orderDailyKpiMap.get(vmsVehicle.getVin());
                    vehicleKpiVo.fillDataForOrderDailyKpi(vmsTosOrderDailyKpis);
                    Map<String, String> map = vehicleKpiVo.toMap();
                    if (CollectionUtils.isNotEmpty(vmsTosOrderDailyKpis)) {
                        int containerNum = vmsTosOrderDailyKpis.stream().map(VmsTosOrderDailyKpi::getContainerIds).filter(StringUtils::isNotEmpty)
                                .flatMap(containerIds -> Arrays.stream(containerIds.split(",")))
                                .collect(Collectors.toSet()).size();
                        map.put("containerNum", String.valueOf(containerNum));
                    } else {
                        map.put("containerNum", "0");
                    }

                    return map;
                } catch (Exception e) {
                    log.error("获取daily kpi异常", e);
                }
                return null;
            }).filter(Objects::nonNull).collect(Collectors.toList());
            return resultList;
        }

        return null;
    }
}
