package com.ssi.service.impl;

import com.alibaba.fastjson.JSON;
import com.ssi.entity.VmsVehicleFreeTime;
import com.ssi.entity.dto.VmsVehicleFreeTimeDto;
import com.ssi.entity.dto.VmsVehicleVinFreeTimeDto;
import com.ssi.service.VmsVehicleFreeTimeService;
import com.ssi.utils.DateUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.elasticsearch.action.search.ClearScrollRequest;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.search.SearchScrollRequest;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.client.indices.GetIndexRequest;
import org.elasticsearch.client.indices.GetIndexResponse;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.MatchQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.RangeQueryBuilder;
import org.elasticsearch.search.Scroll;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.sort.FieldSortBuilder;
import org.elasticsearch.search.sort.SortOrder;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import reactor.util.function.Tuple2;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * @description: 车辆空闲时间 service
 * @author: dt
 * @create: 2022-11-09 18:28
 */
@Service
@Slf4j
public class VmsVehicleFreeTimeServiceImpl implements VmsVehicleFreeTimeService {

    /**
     * es 每个批次查询的条数
     */
    public static final int SEARCH_SIZE = 10000;

    /**
     * es 游标生效时间
     */
    public final Scroll scroll = new Scroll(TimeValue.timeValueMinutes(10L));

    /**
     * es 索引名
     */
    @Value("${vehicle-trip.es-index-name:vms_vehicle_trip}")
    private String tripIndexName;

    /**
     * es
     */
    private RestHighLevelClient restHighLevelClient;

    public VmsVehicleFreeTimeServiceImpl(RestHighLevelClient restHighLevelClient) {
        this.restHighLevelClient = restHighLevelClient;
    }

    /**
     * 查询车辆空闲时间
     * @param vmsVehicleVinFreeTimeDto
     * @return
     */
    @Override
    public VmsVehicleVinFreeTimeDto getVehicleFreeTime(VmsVehicleVinFreeTimeDto vmsVehicleVinFreeTimeDto) {
        if (CollectionUtils.isEmpty(vmsVehicleVinFreeTimeDto.getVmsVehicleFreeTimeDtos())) {
            return null;
        }

        try{
            SearchResponse response = this.getSearchResponse(vmsVehicleVinFreeTimeDto, SEARCH_SIZE, scroll, SortOrder.ASC);
            String scrollId = response.getScrollId();
            SearchHit[] searchHits = response.getHits().getHits();
            while (searchHits != null && searchHits.length > 0) {
                for (SearchHit hit : searchHits) {
                    String sourceAsString = hit.getSourceAsString();
                    if (StringUtils.isNotEmpty(sourceAsString)) {
                        VmsVehicleFreeTime vmsVehicleFreeTime = JSON.parseObject(sourceAsString, VmsVehicleFreeTime.class);
                        vmsVehicleVinFreeTimeDto.getVmsVehicleFreeTimeDtos().forEach(vmsVehicleFreeTimeDto -> {
                            calcFreeTime(vmsVehicleFreeTime, vmsVehicleFreeTimeDto);
                        });
                    }
                }
                if (searchHits.length == SEARCH_SIZE) {
                    response = this.getNextScrollResponse(scrollId, scroll);
                    searchHits = response.getHits().getHits();
                } else {
                    break;
                }

            }
        }catch (Exception e) {
            log.error("查询统计车辆空闲时间出错", e);
        }
        return vmsVehicleVinFreeTimeDto;
    }

    /**
     * 计算空闲时间
     * @param vmsVehicleFreeTime
     * @param vmsVehicleFreeTimeDto
     */
    public void calcFreeTime(VmsVehicleFreeTime vmsVehicleFreeTime, VmsVehicleFreeTimeDto vmsVehicleFreeTimeDto) {

        Long startTime = 0L;
        Long endTime = 0L;

        // 时间不在该范围内
        if (vmsVehicleFreeTime.getStartTime() > vmsVehicleFreeTimeDto.getEndTime() || vmsVehicleFreeTime.getEndTime() < vmsVehicleFreeTimeDto.getStartTime()) {
            return;
        }

        // 确定开始时间
        if (vmsVehicleFreeTime.getStartTime() < vmsVehicleFreeTimeDto.getStartTime()) {
            startTime = vmsVehicleFreeTimeDto.getStartTime();
        }

        if (vmsVehicleFreeTime.getStartTime() > vmsVehicleFreeTimeDto.getStartTime()) {
            startTime = vmsVehicleFreeTime.getStartTime();
        }


        // 确定结束时间
        if (vmsVehicleFreeTime.getEndTime() < vmsVehicleFreeTimeDto.getEndTime()) {
            endTime = vmsVehicleFreeTime.getEndTime();
        }
        if (vmsVehicleFreeTime.getEndTime() > vmsVehicleFreeTimeDto.getEndTime()) {
            endTime = vmsVehicleFreeTimeDto.getEndTime();
        }

        // 计算空闲时间
        vmsVehicleFreeTimeDto.setFreeTime(Objects.nonNull(vmsVehicleFreeTimeDto.getFreeTime()) ? vmsVehicleFreeTimeDto.getFreeTime() + (endTime-startTime) : (endTime-startTime));
    }

    /**
     *
     * @param vmsVehicleVinFreeTimeDto
     * @param size
     * @param scroll
     * @param sortOrder
     * @return
     * @throws IOException
     */
    public SearchResponse getSearchResponse(VmsVehicleVinFreeTimeDto vmsVehicleVinFreeTimeDto, int size, Scroll scroll,
                                            SortOrder sortOrder) throws IOException {
        MatchQueryBuilder matchQueryBuilder = QueryBuilders.matchQuery("vin", vmsVehicleVinFreeTimeDto.getVin());
        BoolQueryBuilder boolBuilder = QueryBuilders.boolQuery();
        boolBuilder.must(matchQueryBuilder);

        vmsVehicleVinFreeTimeDto.getVmsVehicleFreeTimeDtos().forEach(vmsVehicleFreeTimeDto -> {
            RangeQueryBuilder rangeQueryBuilder = QueryBuilders.rangeQuery("startTime");
            rangeQueryBuilder.gte(vmsVehicleFreeTimeDto.getStartTime());
            rangeQueryBuilder.lte(vmsVehicleFreeTimeDto.getEndTime());
            boolBuilder.should(rangeQueryBuilder);
        });

        log.info("查询 es 条件：\n"+ boolBuilder.toString());

        FieldSortBuilder fieldSortBuilder = new FieldSortBuilder("startTime").order(sortOrder);

        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
        sourceBuilder.size(size);
        sourceBuilder.query(boolBuilder);
        sourceBuilder.sort(fieldSortBuilder);
        sourceBuilder.timeout(new TimeValue(600, TimeUnit.SECONDS));
        Map<String, Long> timeMap = getMinStartTimeAndMaxEndTime(vmsVehicleVinFreeTimeDto);
        String[] indexNames = getAllIndexNamesByYear(timeMap.get("startTime"), timeMap.get("endTime"));
        log.info("查询的索引: "+ Arrays.toString(indexNames));
        SearchRequest searchRequest = new SearchRequest(indexNames, sourceBuilder);
        searchRequest.scroll(scroll);

        return restHighLevelClient.search(searchRequest, RequestOptions.DEFAULT);
    }

    /**
     * 判断获取时间最大值和最小值
     * @param vmsVehicleVinFreeTimeDto
     * @return
     */
    private Map<String, Long> getMinStartTimeAndMaxEndTime(VmsVehicleVinFreeTimeDto vmsVehicleVinFreeTimeDto) {
        Map<String, Long> map = new HashMap<String, Long>(2);
        vmsVehicleVinFreeTimeDto.getVmsVehicleFreeTimeDtos().forEach(vmsVehicleFreeTimeDto -> {
            Long startTime = map.getOrDefault("startTime", 0L);
            Long endTime = map.getOrDefault("endTime", 0L);

            if (startTime > vmsVehicleFreeTimeDto.getStartTime()) {
                map.put("startTime", vmsVehicleFreeTimeDto.getStartTime());
            }

            if (endTime < vmsVehicleFreeTimeDto.getEndTime()) {
                map.put("endTime", vmsVehicleFreeTimeDto.getStartTime());
            }
        });
        return map;
    }

    /**
     * 得到下一个游标的数据
     * @param scrollId
     * @param scroll
     * @return
     * @throws IOException
     */
    public SearchResponse getNextScrollResponse(String scrollId, Scroll scroll) throws IOException {
        SearchScrollRequest scrollRequest = new SearchScrollRequest(scrollId);
        scrollRequest.scroll(scroll);

        return restHighLevelClient.scroll(scrollRequest, RequestOptions.DEFAULT);
    }

    /**
     * 清楚游标
     * @param scrollId
     * @throws IOException
     */
    public void clearScroll(String scrollId) throws IOException {
        ClearScrollRequest clearScrollRequest = new ClearScrollRequest();
        clearScrollRequest.addScrollId(scrollId);

        restHighLevelClient.clearScroll(clearScrollRequest, RequestOptions.DEFAULT);
    }

    /**
     * type: 0：year，1：yearmonth
     *
     * @param startTime
     * @param endTime
     * @return
     * @throws IOException
     */
    private String[] getAllIndexNamesByYear(long startTime, long endTime) throws IOException {
        GetIndexRequest request = new GetIndexRequest(tripIndexName + "*");
        request.humanReadable(true);
        GetIndexResponse response = restHighLevelClient.indices().get(request, RequestOptions.DEFAULT);
        String[] indices = response.getIndices();
        int startYear = Integer.parseInt(DateUtils.getYear(startTime));
        int endYear = Integer.parseInt(DateUtils.getYear(endTime));
        List<String> result = new ArrayList<>();
        for (String indexName : indices) {
            String suffix = StringUtils.substringAfter(indexName, tripIndexName + "_");
            if (NumberUtils.isDigits(suffix)) {
                int suffixYearMonth = Integer.parseInt(suffix);
                if (suffixYearMonth >= startYear && suffixYearMonth <= endYear) {
                    result.add(indexName);
                }
            }
        }

        return result.toArray(new String[0]);
    }

}
