package com.ssi.service.impl;


import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.ssi.model.RedisConfigModel;
import com.ssi.service.WebSocketDataBackService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.client.transport.TransportClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;

/**
 * Description:
 *
 * @author LiXiaoCong
 * @version 2019/10/16 17:51
 */
@Service
@Slf4j
public class WebSocketDataBackServiceImpl implements WebSocketDataBackService {

    private static final Object LOCK = new Object();

    @Value("${meilan.history.esIndex:vms_history_data}")
    private String indexBase;

    @Value("${meilan.history.config.group:playBack}")
    private String redisConfigGroup;

    @Value("${meilan.history.playBack.normal:false}")
    private boolean playBackNormal;

    @Value("${meilan.history.playBack.delay:1000}")
    private long delay;

    @Value("${meilan.history.playBack.open.randomDelay:true}")
    private boolean randomDelay;
    @Value("${meilan.history.playBack.open.randomDelayScale:0.05}")
    private double randomDelayScale;

    @Value("${meilan.history.playBack.open.randomDelayScenesId:6,8,9}")
    private String randomDelayScenesId;

    @Value("${LinuxFile.playBackFolder}")
    private String playBackFolder;

    @Autowired
    private RedisConfigModel redisConfigModel;

    @Autowired
    protected TransportClient transportClient;

    @Autowired
    private ExecutorService executorService;

    private Map<String, WebSocketDataEsBackor> backorMap = Maps.newConcurrentMap();

    @Override
    public int playBack(String sessionId,
                        String backorId,
                        String scenesId,
                        Set<String> vins,
                        Long startTime,
                        Long stopTime) {

        synchronized (LOCK) {
            int status;
            if (!backorMap.containsKey(backorId)) {
                if(!backorId.contains("v2x")) {
                    WebSocketDataEsBackor webSocketDataEsBackor = initPlayBackor(backorId, scenesId, vins, sessionId);
                    webSocketDataEsBackor.setRandomDelay(randomDelay);
                    webSocketDataEsBackor.setRandomDelayScale(randomDelayScale);
                    webSocketDataEsBackor.setRandomDelayScenesId(randomDelayScenesId);

                    backorMap.put(backorId, webSocketDataEsBackor);

                    final BackCondition backConditionBean = getBackCondition(scenesId, startTime, stopTime);

                    if (backConditionBean != null) {
                        log.info(String.format("启动回放任务: (%s, %s, %s)", sessionId, backorId, vins));
                        executorService.execute(() -> {
                            webSocketDataEsBackor.playBack(backConditionBean.startTime,
                                    backConditionBean.stopTime, backConditionBean.delay, playBackNormal);
                        });
                        status = 0;
                    } else {
                        status = 1;
                        backorMap.remove(backorId);
                        log.warn(String.format("无回放数据时间段: (%s, %s, %s)", sessionId, backorId, vins));
                    }
                }else{

                    WebSocketDataFileBackor webSocketDataFileBackor = initPlayFileBackor(backorId, scenesId, vins, sessionId);
                    webSocketDataFileBackor.setRandomDelay(randomDelay);
                    webSocketDataFileBackor.setRandomDelayScale(randomDelayScale);
                    webSocketDataFileBackor.setRandomDelayScenesId(randomDelayScenesId);
                    backorMap.put(backorId, webSocketDataFileBackor);
                    final BackCondition backConditionBean = getBackCondition(scenesId, startTime, stopTime);
                    webSocketDataFileBackor.setPath(backConditionBean.getPath());
                    if(StringUtils.isNotBlank(backConditionBean.getVins())){
                        vins.addAll(Arrays.asList(backConditionBean.getVins().split(",")));
                    }
                    if (backConditionBean != null) {
                        log.info(String.format("启动回放任务: (%s, %s, %s)", sessionId, backorId, vins));
                        executorService.execute(() -> {
                            webSocketDataFileBackor.playBack(backConditionBean.startTime,
                                    backConditionBean.stopTime, backConditionBean.delay, playBackNormal);
                        });
                        status = 0;
                    } else {
                        status = 1;
                        backorMap.remove(backorId);
                        log.warn(String.format("无回放数据时间段: (%s, %s, %s)", sessionId, backorId, vins));
                    }
                }
            } else {
                status = 2;
                log.info(String.format("已存在回放任务: (%s, %s, %s)，无需重复启动。", sessionId, backorId, vins));
                backorMap.get(backorId).sessions.add(sessionId);
            }

            return status;
        }
    }

    @Override
    public void close(String sessionId,
                      String backorId) {

        synchronized (LOCK) {
            WebSocketDataEsBackor webSocketDataEsBackor = backorMap.get(backorId);
            if (webSocketDataEsBackor != null) {
                if (sessionId.equals(webSocketDataEsBackor.mainSession)) {
                    webSocketDataEsBackor.shutDown();
                    backorMap.remove(backorId);
                } else {
                    webSocketDataEsBackor.sessions.remove(sessionId);
                }
            }
        }
    }

    private BackCondition getBackCondition(String scenesId,
                                           Long startTime,
                                           Long stopTime){
        BackCondition backCondition = getRedisBackCondition(scenesId);
        if (backCondition == null && startTime != null && stopTime != null) {
            backCondition = new BackCondition(scenesId, startTime, stopTime, this.delay);
        }
        return backCondition;
    }

    private WebSocketDataEsBackor initPlayBackor(String backorId, String scenesId, Set<String> vins, String sessionId) {

        Set<String> sessions = Sets.newConcurrentHashSet();
        sessions.add(sessionId);

        return new WebSocketDataEsBackor(backorId, scenesId, this,
                vins, sessionId, sessions, transportClient, indexBase,playBackFolder);
    }
    private WebSocketDataFileBackor initPlayFileBackor(String backorId, String scenesId, Set<String> vins, String sessionId) {

        Set<String> sessions = Sets.newConcurrentHashSet();
        sessions.add(sessionId);

        return new WebSocketDataFileBackor(backorId, scenesId, this,
                vins, sessionId, sessions);
    }
    private BackCondition getRedisBackCondition(String scenesId){

        BackCondition backCondition = null;
        String config = redisConfigModel.getConfig(redisConfigGroup, scenesId);
        if(config != null){
            String[] backConditions = config.split("#");
            int length = backConditions.length;
            if(length == 2 || length == 3 || length == 5){
                try {
                    long startTime  = Long.parseLong(backConditions[0]);
                    long stopTime   = Long.parseLong(backConditions[1]);

                    Long delay = this.delay;
                    if(length == 3){
                        delay = Long.parseLong(backConditions[2]);
                    }
                    backCondition = new BackCondition(scenesId, startTime, stopTime, delay);

                    if(length == 5 ){
                        backCondition.setPath(backConditions[3]);
                        backCondition.setVins(backConditions[4]);
                    }


                    log.info(String.format("redis配置中场景%s的回放时间配置：%s", scenesId, backCondition));
                } catch (NumberFormatException e) {
                    log.error(String.format("回放条件解析失败：%s", config));
                }
            }else{
                log.error(String.format("redis配置中场景%s的回放时间配置错误：%s", scenesId, config));
            }
        }else{
            log.warn(String.format("redis配置中无此场景id的回放时间配置：%s", scenesId));
        }
        return backCondition;
    }

    private class BackCondition{

        private String scenesId;
        private long startTime;
        private long stopTime;
        private long delay;

        private String path;
        private String vins;
        public BackCondition( String scenesId, long startTime, long stopTime, long delay){
            this.scenesId = scenesId;
            this.startTime = startTime;
            this.stopTime = stopTime;
            this.delay = delay;
        }

        @Override
        public String toString() {
            return String.format("(%s, %s, %s, %s)", scenesId, startTime, stopTime, delay);
        }

        public String getScenesId() {
            return scenesId;
        }

        public void setScenesId(String scenesId) {
            this.scenesId = scenesId;
        }

        public long getStartTime() {
            return startTime;
        }

        public void setStartTime(long startTime) {
            this.startTime = startTime;
        }

        public long getStopTime() {
            return stopTime;
        }

        public void setStopTime(long stopTime) {
            this.stopTime = stopTime;
        }

        public long getDelay() {
            return delay;
        }

        public void setDelay(long delay) {
            this.delay = delay;
        }

        public String getPath() {
            return path;
        }

        public void setPath(String path) {
            this.path = path;
        }

        public String getVins() {
            return vins;
        }

        public void setVins(String vins) {
            this.vins = vins;
        }
    }

}
