package com.ssi.task;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.ssi.entity.VmsChargingOrderInfo;
import com.ssi.entity.VmsChargingStation;
import com.ssi.entity.VmsChargingTotal;
import com.ssi.mapper.VmsChargingTotalMapper;
import com.ssi.model.RedisDataModel;
import com.ssi.service.VmsChargingConnectorInfoService;
import com.ssi.service.VmsChargingEquipmentInfoService;
import com.ssi.service.VmsChargingOrderInfoService;
import com.ssi.service.VmsChargingStationService;
import com.ssi.utils.telaidian.TelaidianRequestTpl;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.mvel2.util.Make;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

/**
 * @author SunnyHotz
 * @PackageName:com.ssi.task
 * @ClassName:TELDStationSyncTask
 * @Description: 特来电充电桩信息及状态同步
 * @date 2022/9/1 13:54
 */
@Slf4j
@Component
public class TELDStationSyncTask {

    private static final ExecutorService executor = Executors.newFixedThreadPool(3);

    @Autowired
    private TelaidianRequestTpl telaidianRequestTpl;
    @Autowired
    private VmsChargingStationService vmsChargingStationService;
    @Autowired
    private VmsChargingEquipmentInfoService chargingEquipmentInfoService;
    @Autowired
    private VmsChargingConnectorInfoService chargingConnectorInfoService;
    @Autowired
    private VmsChargingOrderInfoService chargingOrderInfoService;
    @Autowired
    private VmsChargingTotalMapper chargingTotalMapper;
    @Autowired
    private RedisDataModel redisDataModel;

    /**
     * 充电桩信息维护
     */
    @Scheduled(cron="0 0 0 * * ?")
    public void syncEquipmentInfo(){

        try {
            JSONObject qryObj = new JSONObject();
            qryObj.put("LastQueryTime", "");
            qryObj.put("PageNo", 1);
            qryObj.put("PageSize", 100);
            JSONObject queryResult = telaidianRequestTpl.queryData(qryObj.toJSONString(), "query_stations_info", 1);
            if (queryResult.getInteger("Ret") != 0) {
                //请求异常信息，结束
                log.error("TLD充电站信息请求异常：{}", queryResult.toJSONString());
                return;
            }
            JSONObject data = queryResult.getJSONObject("Data");
            //更新充电站信息
            CompletableFuture<Void> future = CompletableFuture.runAsync(vmsChargingStationService.saveStationInfo(data), executor);
            //更新设备信息
            CompletableFuture<Void> future1 = CompletableFuture.runAsync(chargingEquipmentInfoService.saveEquipmentInfo(data), executor);
            //更新充电接口信息
            CompletableFuture<Void> future2 = CompletableFuture.runAsync(chargingConnectorInfoService.saveConnectorInfo(data), executor);
            CompletableFuture<Void>  allOfFuture = CompletableFuture.allOf(future, future1,future2);
            allOfFuture.get();
        }catch (Exception e){
            log.error("syncEquipmentInfo occur error:",e);
        }
    }

    /**
     * 定时同步设备接口状态
     */
    @Scheduled(cron="0 */5 * * * ?")
    public void syncConnectorStatus(){

        try{
            List<VmsChargingStation> list = vmsChargingStationService.list();
            if(CollectionUtils.isEmpty(list)){
                return;
            }
            List<String> collect = list.stream().map(VmsChargingStation::getStationId).collect(Collectors.toList());

            JSONObject qryObj = new JSONObject();
            qryObj.put("StationIDs", collect);
            JSONObject queryResult = telaidianRequestTpl.queryData(qryObj.toJSONString(), "query_station_status", 1);
            if (queryResult.getInteger("Ret") != 0) {
                //请求异常信息，结束
                log.error("TLD充电站信息请求异常：{}", queryResult.toJSONString());
                return;
            }
            JSONObject data = queryResult.getJSONObject("Data");
            JSONArray stationStatusInfos = data.getJSONArray("StationStatusInfos");
            for (int i = 0; i < stationStatusInfos.size(); i++) {
                chargingConnectorInfoService.updateStatus(stationStatusInfos.getJSONObject(i));
            }
        }catch (Exception e){
            log.error("syncConnectorStatus occur error:",e);
        }
    }

    /**
     * 查询充电站订单信息
     */
    @Scheduled(cron="0 */10 * * * ?")
    public void qryCurrentStationOrder() {

        try{
            List<VmsChargingStation> list = vmsChargingStationService.list();
            if(CollectionUtils.isEmpty(list)){
                return;
            }
            List<String> collect = list.stream().map(VmsChargingStation::getStationId).collect(Collectors.toList());
            for (String stationID: collect){
                JSONObject qryObj = new JSONObject();
                qryObj.put("StationID", stationID);
                JSONObject queryResult = telaidianRequestTpl.queryData(qryObj.toJSONString(), "query_chargingorder_list", 1);
                if (queryResult.getInteger("Ret") != 0) {
                    //请求异常信息，结束
                    log.error("TLD充电站当前订单请求异常：{}", queryResult.toJSONString());
                    return;
                }
                JSONObject data = queryResult.getJSONObject("Data");
                JSONArray orderCodeS = data.getJSONArray("OrderCodeS");
                for (int i = 0; i < orderCodeS.size(); i++) {
                    String orderId = orderCodeS.getString(i);
                    VmsChargingOrderInfo byId = chargingOrderInfoService.getById(orderId);
                    if(Objects.isNull(byId)){
                        VmsChargingOrderInfo vmsChargingOrderInfo = componentChargingOrderInfo(orderId);
                        chargingOrderInfoService.save(vmsChargingOrderInfo);
                    }
                }
            }
        }catch (Exception e){
            log.error("qryCurrentStationOrder occur error:",e);
        }

    }

    /**
     * 查询已完成充电站订单信息
     */
    @Scheduled(cron="0 */30 * * * ?")
    public void collectCompeletedOrders(){

        try{
            JSONObject qryObj = new JSONObject();
            DateTimeFormatter df = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
            LocalDateTime now  = LocalDateTime.now();
            LocalDateTime beforeTime = now.minusHours(2);

            qryObj.put("QueryStartTime", df.format(beforeTime));
            qryObj.put("QueryEndTime", df.format(now));
            JSONObject queryResult = telaidianRequestTpl.queryData(qryObj.toJSONString(), "query_order_info", 1);
            if (queryResult.getInteger("Ret") != 0) {
                //请求异常信息，结束
                log.error("TLD充电站已完成订单请求异常：{}", queryResult.toJSONString());
                return;
            }
            JSONObject data = queryResult.getJSONObject("Data");
            JSONArray orderInfos = data.getJSONArray("OrderInfos");
            for (int i = 0; i < orderInfos.size(); i++) {
                JSONObject orderObj = orderInfos.getJSONObject(i);
                VmsChargingOrderInfo oldEntity = chargingOrderInfoService.getById(orderObj.getString("StartChargeSeq"));
                if(Objects.isNull(oldEntity)){
                    oldEntity =  new VmsChargingOrderInfo();
                    oldEntity.setStartChargeSeq(orderObj.getString("StartChargeSeq"));
                }
                oldEntity.setElect(orderObj.getDouble("Elect"));
                oldEntity.setStartTime(orderObj.getDate("StartTime"));
                oldEntity.setEndTime(orderObj.getDate("EndTime"));
                chargingOrderInfoService.saveOrUpdate(oldEntity);
//                if(StringUtils.isNotBlank(oldEntity.getVin())){
//                    saveChargingTotal(orderObj,oldEntity.getVin());
//                }
            }
        }catch (Exception e){
            log.error("collectCompeletedOrders occur error:",e);
        }
    }

    private void saveChargingTotal(JSONObject oldEntity,String vin) {

        VmsChargingTotal startChargeSeq = chargingTotalMapper.selectById(oldEntity.getString("StartChargeSeq"));
        VmsChargingTotal vmsChargingTotal = new VmsChargingTotal();
        vmsChargingTotal.setId(oldEntity.getString("StartChargeSeq"));
        vmsChargingTotal.setChargingEnergy(oldEntity.getBigDecimal("Elect"));
        vmsChargingTotal.setYear(String.valueOf(LocalDate.now().getYear()));
        vmsChargingTotal.setVin(vin);
        vmsChargingTotal.setStartTime(oldEntity.getDate("StartTime").getTime());
        vmsChargingTotal.setEndTime(oldEntity.getDate("EndTime").getTime());
        vmsChargingTotal.setDuration(vmsChargingTotal.getEndTime()-vmsChargingTotal.getStartTime());
        if(Objects.isNull(startChargeSeq)){
            chargingTotalMapper.insert(vmsChargingTotal);
        }else {
            chargingTotalMapper.updateById(vmsChargingTotal);
        }
    }


    /**
     * 请求当前订单信息
     * @param orderId
     * @return
     */
    private VmsChargingOrderInfo componentChargingOrderInfo(String orderId) {
        if(StringUtils.isBlank(orderId)){
            return null;
        }
        VmsChargingOrderInfo defaultResult = new VmsChargingOrderInfo();
        defaultResult.setStartChargeSeq(orderId);
        try{
            JSONObject qryObj = new JSONObject();
            qryObj.put("StartChargeSeq", orderId);
            JSONObject queryResult = telaidianRequestTpl.queryData(qryObj.toJSONString(), "query_equip_charge_status", 1);
            if (queryResult.getInteger("Ret") != 0) {
                //请求异常信息，结束
                log.error("TLD充电站订单详细信息请求异常：{}", queryResult.toJSONString());
                return defaultResult;
            }
            String data = queryResult.getString("Data");
            if(StringUtils.isNotBlank(data)){
                VmsChargingOrderInfo vmsChargingOrderInfo = JSONObject.parseObject(data, VmsChargingOrderInfo.class);
                if(StringUtils.isNotBlank(vmsChargingOrderInfo.getVin())){
                    String s = redisDataModel.get("harbor:command:status:".concat(vmsChargingOrderInfo.getVin()));
                    JSONObject jsonObject = JSON.parseObject(s);
                    if(jsonObject.getInteger("chargingStatus")==1){
                        vmsChargingOrderInfo.setTaskNo(jsonObject.getString("taskNo"));
                    }
                }
                vmsChargingOrderInfo.setStartChargeSeq(orderId);
                return vmsChargingOrderInfo;
            }
        }catch (Exception e){
            log.error("componentChargingOrderInfo occur error:",e);
        }
        return defaultResult;
    }

}
