package com.ssi.utils;

import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.admin.indices.exists.indices.IndicesExistsRequest;
import org.elasticsearch.action.admin.indices.exists.indices.IndicesExistsResponse;
import org.elasticsearch.action.admin.indices.exists.types.TypesExistsRequest;
import org.elasticsearch.action.admin.indices.exists.types.TypesExistsResponse;
import org.elasticsearch.client.transport.TransportClient;
import org.elasticsearch.cluster.metadata.IndexMetaData;
import org.elasticsearch.common.collect.ImmutableOpenMap;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.RangeQueryBuilder;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.sort.FieldSortBuilder;
import org.elasticsearch.search.sort.SortOrder;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.joda.time.Days;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.elasticsearch.core.ElasticsearchTemplate;
import org.springframework.data.elasticsearch.core.query.NativeSearchQueryBuilder;
import org.springframework.data.elasticsearch.core.query.SearchQuery;

import java.util.*;

import static org.elasticsearch.index.query.QueryBuilders.boolQuery;

/**
 * Description:
 *
 * @author LiXiaoCong
 * @version 2018/4/27 8:30
 */
@Slf4j
public class ElasticSearchUtil {

  private final static DateTimeZone TIME_ZONE = DateTimeZone.forID("Asia/Shanghai");


  private ElasticSearchUtil() {
  }

  public static boolean isExistsIndex(TransportClient client,
      String indexName) {
    if (StringUtils.isNotBlank(indexName)) {
      IndicesExistsResponse response =
          client.admin().indices().exists(
              new IndicesExistsRequest().indices(new String[]{indexName})).actionGet();
      return response.isExists();
    }
    return false;
  }


  public static boolean isExistsType(TransportClient client,
      String indexName,
      String indexType) {

    if (StringUtils.isNotBlank(indexName) && StringUtils.isNotBlank(indexType)) {
      TypesExistsResponse response =
          client.admin().indices()
              .typesExists(new TypesExistsRequest(
                  new String[]{indexName}, indexType)).actionGet();
      return response.isExists();
    }
    return false;
  }

  public static List<String> getIndices(TransportClient client) {
    ImmutableOpenMap<String, IndexMetaData> indexMap = client.admin()
        .cluster().prepareState().execute().actionGet().getState().getMetaData().getIndices();
    Iterator<String> keysIt = indexMap.keysIt();
    return Lists.newArrayList(keysIt);
  }

  public static List<Map<String, Object>> selectItem(SearchHits searchHits) {
    SearchHit[] hits = searchHits.getHits();
    List<Map<String, Object>> resList = Lists.newLinkedList();
    for (SearchHit hit : hits) {
      resList.add(hit.getSourceAsMap());
    }
    return resList;
  }

  public static List<Map<String, Object>> selectItemWithLocationTypeChange(SearchHits searchHits,
      String toType) {
    SearchHit[] hits = searchHits.getHits();
    List<Map<String, Object>> resList = Lists.newLinkedList();
    Map<String, Object> sourceAsMap;
    for (SearchHit hit : hits) {
      sourceAsMap = hit.getSourceAsMap();
      locationTypeChange(sourceAsMap, toType);
      resList.add(sourceAsMap);
    }
    return resList;
  }

  public static RangeQueryBuilder createRangeQueryBuilder(String field,
      Long startTime,
      Long endTime) {
    RangeQueryBuilder rangeQueryBuilder = null;
    if (startTime != null || endTime != null) {
      rangeQueryBuilder = QueryBuilders.rangeQuery(field);
      if (startTime != null) {
        rangeQueryBuilder.gte(startTime);
      }
      if (endTime != null) {
        rangeQueryBuilder.lte(endTime);
      }

    }
    return rangeQueryBuilder;
  }


  /**
   * 将map中的经纬度 进行高德地图纠偏
   */
  public static void locationTypeChange(Map<String, Object> sourceAsMap, String toType) {
    if (sourceAsMap != null && toType != null) {
      Object longitudeObj = sourceAsMap.get("longitude");
      Object latitudeObj = sourceAsMap.get("latitude");
      if (latitudeObj != null && longitudeObj != null) {

        double longitude = (double) longitudeObj;
        double latitude = (double) latitudeObj;

        double[] doubles = null;
        switch (toType) {
          //高德
          case "gcj02":
            doubles = GeoPosTransformUtil.wgs84togcj02(longitude, latitude);
            break;
          //百度
          case "bd09":
            doubles = GeoPosTransformUtil.wgs84tobd09(longitude, latitude);
            break;
        }

        if (doubles != null) {
          sourceAsMap.put("longitude", doubles[0]);
          sourceAsMap.put("latitude", doubles[1]);
        }
      }
    }
  }

  public static boolean chinaLocationTypeChange(Map<String, Object> sourceAsMap, String toType) {
    boolean status = true;
    if (sourceAsMap != null && toType != null) {
      Object longitudeObj = sourceAsMap.get("longitude");
      Object latitudeObj = sourceAsMap.get("latitude");
      if (latitudeObj != null && longitudeObj != null) {

        double longitude = (double) longitudeObj;
        double latitude = (double) latitudeObj;

        boolean inChina = isInChina(longitude, latitude);
        if (inChina) {
          double[] doubles = null;
          switch (toType) {
            //高德
            case "gcj02":
              doubles = GeoPosTransformUtil.wgs84togcj02(longitude, latitude);
              break;
            //百度
            case "bd09":
              doubles = GeoPosTransformUtil.wgs84tobd09(longitude, latitude);
              break;
          }

          if (doubles != null) {
            sourceAsMap.put("longitude", doubles[0]);
            sourceAsMap.put("latitude", doubles[1]);
          }
        } else {
          status = false;
        }
      }
    }
    return status;
  }

  public static boolean isInChina(double longitude, double latitude) {
    return (longitude >= 73 && longitude <= 135 && latitude >= 4 && latitude <= 53);
  }

  /**
   * 构建搜索查询
   */
  public static SearchQuery buildSearchQuery(ElasticsearchTemplate elasticsearchTemplate,
      List<String> indexList, String typeName, String[] searchFields, String sortField,
      SortOrder order, BoolQueryBuilder boolQueryBuilder, int page, int pageSize) {
    String[] indexArray = indexList.stream().filter(elasticsearchTemplate::indexExists)
        .toArray(String[]::new);
    NativeSearchQueryBuilder nativeSearchQueryBuilder = new NativeSearchQueryBuilder()
        .withIndices(indexArray)
        .withTypes(typeName)
        .withQuery(boolQueryBuilder)
        .withFields(searchFields)
        .withSort(new FieldSortBuilder(sortField).order(order));
    SearchQuery searchQuery = null;
    if (pageSize == 0) {
      searchQuery = nativeSearchQueryBuilder.build();
    } else {
      searchQuery = nativeSearchQueryBuilder
          .withPageable(PageRequest.of(page, pageSize))
          .build();
    }
    return searchQuery;
  }

  /**
   * 查询语句构建
   *
   * @param termMap 单值精确查询参数
   * @param termsMap 多值精确查询参数
   * @param longRangeMap long类型范围查询参数
   * @param doubleRangeMap double类型范围查询参数
   */
  public static BoolQueryBuilder getQueryBuilder(Map<String, String> termMap,
      Map<String, List<Integer>> termsMap, String timeField, long startTime, long endTime,
      Map<String, Long[]> longRangeMap, Map<String, Double[]> doubleRangeMap) {
    BoolQueryBuilder boolQueryBuilder = boolQuery();
    if (termMap != null && termMap.size() > 0) {
      for (Map.Entry<String, String> entry : termMap.entrySet()) {
        boolQueryBuilder.must(QueryBuilders.termQuery(entry.getKey(), entry.getValue()));
      }
    }

    if (termsMap != null && termsMap.size() > 0) {
      for (Map.Entry<String, List<Integer>> entry : termsMap.entrySet()) {
        boolQueryBuilder.must(QueryBuilders.termsQuery(entry.getKey(), entry.getValue()));
      }
    }
    Long[] timeRange = {startTime, endTime};
    if (longRangeMap == null) {
      longRangeMap = new HashMap<String, Long[]>();
    }
    longRangeMap.put(timeField, timeRange);

    if (longRangeMap != null && longRangeMap.size() > 0) {
      for (Map.Entry<String, Long[]> entry : longRangeMap.entrySet()) {
        boolQueryBuilder.must(QueryBuilders.rangeQuery(entry.getKey())
            .gte(entry.getValue()[0])
            .lte(entry.getValue()[1]));
      }
    }

    if (doubleRangeMap != null && doubleRangeMap.size() > 0) {
      for (Map.Entry<String, Double[]> entry : doubleRangeMap.entrySet()) {
        boolQueryBuilder.must(QueryBuilders.rangeQuery(entry.getKey())
            .gte(entry.getValue()[0])
            .lte(entry.getValue()[1]));
      }
    }

    return boolQueryBuilder;
  }

  public static BoolQueryBuilder getQueryBuilder(String key, String value, String timeField,
      long startTime, long endTime) {
    Map<String, String> termMap = new HashMap<>();
    termMap.put(key, value);
    Map<String, Double[]> doubleRangeMap = new HashMap<String, Double[]>();
    Double[] latLonRange = {1d, 360d};
    doubleRangeMap.put("latitude", latLonRange);
    doubleRangeMap.put("longitude", latLonRange);
    return getQueryBuilder(termMap, null, timeField, startTime, endTime, null, doubleRangeMap);
  }

  public static String[] getIndexs(Long startTime,
      Long stopTime,
      String indexBase,
      int defaultDayPeriod,
      boolean defaultContainsToday,
      String indexRule) {

    if (indexRule == null) {
      return new String[]{indexBase};
    }

    DateTime stopDate;
    if (stopTime == null) {
      DateTime now = DateTime.now(TIME_ZONE);
      if (defaultContainsToday) {
        stopDate = now;
      } else {
        stopDate = now.minusDays(1);
      }
    } else {
      stopDate = new DateTime(stopTime.longValue(), TIME_ZONE);
    }

    DateTime startDate = (startTime == null ? stopDate.minusDays(defaultDayPeriod)
        : new DateTime(startTime.longValue(), TIME_ZONE));

    List<String> indexList = new ArrayList<>();
    String[] indexs;
    int days = Days.daysBetween(startDate, stopDate).getDays();
    Set<String> indexSet = new HashSet<>();
    if (days >= 0) {
      for (int i = 0; i <= days; i++) {
        try {
          switch (indexRule.toLowerCase()) {
            //按年分区
            case "year":
              indexSet.add(String.format("%s_%s*",
                  indexBase, startDate.plusDays(i).toString("yyyy")));
              break;
            //按月分区
            case "month":
              indexSet.add(String.format("%s_%s*",
                  indexBase, startDate.plusDays(i).toString("yyyyMM")));
              break;
            //按天分区
            case "day":
              indexList.add(String.format("%s_%s*",
                  indexBase, startDate.plusDays(i).toString("yyyyMMdd")));
              break;
            default:
              indexSet.add(indexBase);
          }
        } catch (Exception e) {
        }
      }
      indexList.addAll(indexSet);
      indexs = indexList.toArray(new String[indexList.size()]);
    } else {
      indexs = new String[0];
    }

    log.info("查询索引有：indexs = " + Lists.newArrayList(indexs).toString());

    return indexs;
  }

}
