package com.ssi.utils.grid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 *  网格对象：
 *      包含网格标识点MercatorPoint，订阅用户队列、当前占用用户
 *      后续考虑采用触发器的形式 在让给下一个用户时直接告知下一个用户可以开始操作了
 *
 */
public class MercatorGrid implements Serializable {

    private final Logger logger = LoggerFactory.getLogger(getClass());

    private final MercatorPoint point;

    //用户占用等待队列
    private final ConcurrentLinkedQueue<String> userWaitQueue;

    //当前用户
    private volatile String currentUser;

    public MercatorGrid(MercatorPoint point){
        assert point != null : "网格标识点不能为空！";
        this.point = point;
        this.userWaitQueue =  new ConcurrentLinkedQueue<>();
    }


    /**
     * 预定此网格
     * @param user  id
     * @return
     */
    public boolean reserve(String user){
        String currentUser = getCurrentUser();
        if(!user.equals(currentUser)) {
            boolean offer = this.userWaitQueue.offer(user);
            //当前用户为null时，设置为此用户
            getCurrentUser();
            logger.debug(String.format("%s 预定占用网格 %s 成功，当前等待用户数：%s",
                    user, point, userWaitQueue.size()));
            return offer;
        }
        return true;
    }

    public boolean cancelReserve(String user){
        this.userWaitQueue.remove(user);
        logger.debug(String.format("取消%s 预定占用网格 %s 成功，当前等待用户数：%s",
                user, point, userWaitQueue.size()));
        return false;
    }

    /**
     * 获取当前用户
     * @return
     */
    public String getCurrentUser(){
        if(this.currentUser == null){
            synchronized (MercatorGrid.class) {
                if(this.currentUser == null) {
                    this.currentUser = this.userWaitQueue.poll();
                }
            }
        }
        return this.currentUser;
    }

    /**
     * 释放当前用户, 让给下一个用户
     * @param currentUser
     * @return
     */
    public boolean freeToNext(String currentUser){
        if(currentUser.equals(this.currentUser)){
            synchronized (MercatorGrid.class){
                if(currentUser.equals(this.currentUser)){
                    this.currentUser = this.userWaitQueue.poll();
                    logger.debug(String.format("%s 释放网格 %s 成功，当前用户变更为：%s, 目前等待用户数",
                            currentUser, point, this.currentUser, userWaitQueue.size()));
                    return true;
                }
            }
        }
        logger.error(String.format("%s 非网格 %s 当前用户 %s，无法释放!", currentUser, point, this.currentUser));
        return false;
    }

    public MercatorPoint getPoint() {
        return this.point.clone();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        MercatorGrid that = (MercatorGrid) o;

        return point.equals(that.point);
    }

    @Override
    public int hashCode() {
        return point != null ? point.hashCode() : 0;
    }

    @Override
    public String toString() {
        final StringBuffer sb = new StringBuffer("MercatorGrid{");
        sb.append("point=").append(point);
        sb.append(", userWaitQueueSize=").append(userWaitQueue.size());
        sb.append(", currentUser='").append(currentUser).append('\'');
        sb.append('}');
        return sb.toString();
    }
}
