package com.ssi.utils.grid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.util.*;

/**
 *   四个顶点，按指定距离划分格子，生成网
 * @author jechedo
 * @time   2022/12/1
 */
public class MercatorNet implements Serializable {

    private final Logger logger = LoggerFactory.getLogger(getClass());

    private final MercatorPoint pointA;
    private final MercatorPoint pointB;
    private final MercatorPoint pointC;
    private final MercatorPoint pointD;

    private final int gridDistance;

    private final Map<MercatorPoint, MercatorGrid> gridsMap;

    /**
     * @param pointA
     * @param pointB
     * @param pointC
     * @param pointD
     * @param gridDistance   网格距离， 单位：米
     */
    public MercatorNet(MercatorPoint pointA,
                       MercatorPoint pointB,
                       MercatorPoint pointC,
                       MercatorPoint pointD,
                       int gridDistance){

        assert gridDistance > 0 : "gridDistance 必须为大于0的值。";

        this.pointA = pointA;
        this.pointB = pointB;
        this.pointC = pointC;
        this.pointD = pointD;
        this.gridDistance = gridDistance;
        this.gridsMap = new HashMap<>();

        //生成网格
        createNet();
    }

    /**
     * 生成网格
     *
     */
    private void createNet(){
        this.gridsMap.putAll(searchGrids(this.pointA, this.pointB, this.pointC, this.pointD));
    }

    private Map<MercatorPoint, MercatorGrid> searchGrids(MercatorPoint pointA,
                                                        MercatorPoint pointB,
                                                        MercatorPoint pointC,
                                                        MercatorPoint pointD){
        //根据四个顶点坐标，求网格x与y的最大最小值
        List<Double> xs = new ArrayList<>();
        xs.add(pointA.getX());
        xs.add(pointB.getX());
        xs.add(pointC.getX());
        xs.add(pointD.getX());
        //向上取整， 获取能整除gridDistance 且距离x最大值最近 又大于x最大值的值
        double maxX = Math.ceil(Collections.max(xs) / gridDistance) * gridDistance;
        //向下取整， 获取能整除gridDistance 且距离x最小值最近 又小于x最小值的值
        double  minX = Math.floor(Collections.min(xs) / gridDistance) * gridDistance;

        List<Double> ys = new ArrayList<>();
        ys.add(pointA.getY());
        ys.add(pointB.getY());
        ys.add(pointC.getY());
        ys.add(pointD.getY());
        //向上取整， 获取能整除gridDistance 且距离y最大值最近 又大于y最大值的值
        double maxY = Math.ceil(Collections.max(ys) / gridDistance) * gridDistance;
        //向下取整， 获取能整除gridDistance 且距离y最小值最近 又小于y最小值的值
        double minY = Math.floor(Collections.min(ys) / gridDistance) * gridDistance;

        return searchGrids(maxX, minX, maxY,minY);
    }

    private Map<MercatorPoint, MercatorGrid> searchGrids(double maxX, double minX, double maxY, double minY){
        Map<MercatorPoint, MercatorGrid> gridsMap = new HashMap<>();
        //x方向 网格个数
        double gridXNum = (maxX - minX) / this.gridDistance;
        //x方向 网格个数
        double gridYNum = (maxY - minY) / this.gridDistance;

        //生成网格, 取网格左下角的点的坐标作为网格标识
        double x;
        double y;
        MercatorPoint point;
        for (int i = 0; i < gridXNum; i++) {
            x  = minX + this.gridDistance * i;
            for (int j = 0; j < gridYNum; j++) {
                y = minY + this.gridDistance * j;
                point = new MercatorPoint(x, y);
                //设置网格状态
                gridsMap.put(point, getMercatorGrid(point));
            }
        }
        return gridsMap;
    }

    public int getNetGridSize(){
        return this.gridsMap.size();
    }

    public Collection<MercatorGrid> getChildGrids(MercatorPoint pointA,
                                                  MercatorPoint pointB,
                                                  MercatorPoint pointC,
                                                  MercatorPoint pointD){


       return searchGrids(pointA, pointB, pointC,pointD).values();
    }

    public Collection<MercatorGrid> getChildGrids(double maxX, double minX, double maxY, double minY){
        maxX = Math.ceil(maxX / gridDistance) * gridDistance;
        minX = Math.floor(minX / gridDistance) * gridDistance;

        maxY = Math.ceil(maxY / gridDistance) * gridDistance;
        minY = Math.floor(minY / gridDistance) * gridDistance;

        return searchGrids(maxX, minX, maxY, minY).values();
    }

    public MercatorGrid getMercatorGrid(MercatorPoint point){
        return this.gridsMap.getOrDefault(point, new MercatorGrid(point));
    }

    public static void main(String[] args) {

        double[] a = Coordtransform.WGS84toMercator(117.964504, 24.454129);
        MercatorPoint pointA = new MercatorPoint(a[0], a[1]);

        double[] b = Coordtransform.WGS84toMercator(117.98918, 24.449011);
        MercatorPoint pointB = new MercatorPoint(b[0], b[1]);

        double[] c = Coordtransform.WGS84toMercator(117.990382, 24.453953);
        MercatorPoint pointC = new MercatorPoint(c[0], c[1]);

        double[] d = Coordtransform.WGS84toMercator(117.965813, 24.458992);
        MercatorPoint pointD = new MercatorPoint(d[0], d[1]);

        MercatorNet mercatorNet = new MercatorNet(pointA, pointB, pointC, pointD, 5);
        Collection<MercatorGrid> grids = mercatorNet.getChildGrids(1.3132610797736611E7, 1.3132030377911612E7, 2809111.607682695, 2808734.2196206213);

       grids.forEach((grid -> {
            grid.reserve("912");
            System.out.println(mercatorNet.getMercatorGrid(grid.getPoint()) + " -- " + grid);
        }));

        System.out.println(mercatorNet.getNetGridSize() + " -- " + grids.size());



      /*  System.out.println( " ---" );
        double[] aa = Coordtransform.WGS84toMercator(117.967487, 24.456316);
        MercatorPoint pointAA = new MercatorPoint(aa[0], aa[1]);
        System.out.println(pointAA);

        double[] bb = Coordtransform.WGS84toMercator(117.97225, 24.455418);
        MercatorPoint pointBB = new MercatorPoint(bb[0], bb[1]);
        System.out.println(pointBB);

        double[] cc = Coordtransform.WGS84toMercator(117.971735, 24.45323);
        MercatorPoint pointCC = new MercatorPoint(cc[0], cc[1]);
        System.out.println(pointCC);

        double[] dd = Coordtransform.WGS84toMercator(117.967036, 24.454148);
        MercatorPoint pointDD = new MercatorPoint(dd[0], dd[1]);
        System.out.println(pointDD);*/

    }

}
