package com.ssi.utils.grid;

import com.google.common.eventbus.EventBus;
import com.google.common.eventbus.Subscribe;
import com.ssi.entity.dto.Point;
import com.ssi.kafka.listener.VehicleRealTimeInfoListener;
import com.ssi.utils.FieldUtil;
import com.ssi.utils.SpringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

public class RealTimeVehicleSchedule {

    private final Logger logger = LoggerFactory.getLogger(getClass());

    private volatile static RealTimeVehicleSchedule VEHICLE_SCHEDULE;

    /** 地图角度修正系数,原北向为0，实际东向为0, 逆时针为正*/
    private static final double deltAngle = 450.0;

    private final MercatorNet mercatorNet;

    private final Map<String, MercatorVehicle> vehicleMap;

    private final EventBus eventBus;

    private RealTimeVehicleSchedule(){
        this.mercatorNet = createTestNet();
        this.vehicleMap = new HashMap<>();

        this.eventBus = new EventBus("GridFree");
        //注册监听到eventBus
        this.eventBus.register(this);
    }


    private MercatorNet  createTestNet(){
        //测试用  从高德地图中拾取的厦门远海的四个顶点坐标
        double[] wgs84 = Coordtransform.GCJ02toWGS84(117.964504, 24.454129);
        double[] a = Coordtransform.WGS84toMercator(wgs84[0], wgs84[1]);
        MercatorPoint pointA = new MercatorPoint(a[0], a[1]);

        wgs84 = Coordtransform.GCJ02toWGS84(117.98918, 24.449011);
        double[] b = Coordtransform.WGS84toMercator(wgs84[0], wgs84[1]);
        MercatorPoint pointB = new MercatorPoint(b[0], b[1]);

        wgs84 = Coordtransform.GCJ02toWGS84(117.990382, 24.453953);
        double[] c = Coordtransform.WGS84toMercator(wgs84[0], wgs84[1]);
        MercatorPoint pointC = new MercatorPoint(c[0], c[1]);

        wgs84 = Coordtransform.GCJ02toWGS84(117.965813, 24.458992);
        double[] d = Coordtransform.WGS84toMercator(wgs84[0], wgs84[1]);
        MercatorPoint pointD = new MercatorPoint(d[0], d[1]);

        return new MercatorNet(pointA, pointB, pointC, pointD, 2);
    }

    public static RealTimeVehicleSchedule getInstance(){
        if(VEHICLE_SCHEDULE == null) {
            synchronized (RealTimeVehicleSchedule.class){
                if(VEHICLE_SCHEDULE == null){
                    VEHICLE_SCHEDULE = new RealTimeVehicleSchedule();
                }
            }
        }
        return VEHICLE_SCHEDULE;
    }


    /**
     * 车辆实时数据
     * @param vehicleDataMap    key：车辆唯一标识， value：实时数据
     *                          其中实时数据需包含字段有：location[longitude、latitude]、speed、direction, collectTime,isOnline
     */
    public void pushRealTimeVehicleData(Map<String, Map<String, Object>> vehicleDataMap){
        if(vehicleDataMap != null){
            Set<Map.Entry<String, Map<String, Object>>> entrySet = vehicleDataMap.entrySet();
            String vid;
            Map<String, Object> data;
            for (Map.Entry<String, Map<String, Object>> entry : entrySet) {
                data = entry.getValue();
                vid = String.format("%s-%s", data.get("vehicleNum"), entry.getKey());
                //处理9子开头的车，并排除950
                if( !vid.startsWith("9") || vid.startsWith("950") || !data.containsKey("location")){
                    continue;
                }
                updateOrCreateVehicle(vid, data);
            }
        }
    }

    /**
     *
     * @param vehicleDataMap
     */
    public void pushSingleVehicleData(Map<String, Object> vehicleDataMap){
        if(vehicleDataMap != null){
                String vid = String.format("%s-%s", vehicleDataMap.get("vehicleNum"), vehicleDataMap.get("vin"));
                //处理9子开头的车，并排除950
                if( !vid.startsWith("9") || vid.startsWith("950") || !vehicleDataMap.containsKey("location")){
                    return;
                }
                updateOrCreateVehicle(vid, vehicleDataMap);
        }

    }

    private void updateOrCreateVehicle(String vid, Map<String, Object> data){
        Object collectTime = data.get("collectTime");
        Object direction = data.get("direction");
        Object speed = data.get("speed");
        data.put("closeArea","");
        String isOnline = String.valueOf(data.get("isOnline"));
        Integer versionNo = (Integer)data.get("versionNo");
        try {
            VehicleRealTimeInfoListener listener = SpringUtils.getObject(VehicleRealTimeInfoListener.class);
            double sp = Double.parseDouble(String.valueOf(speed));
            double[] location = (double[]) data.get("location");
            double direc = Double.parseDouble(String.valueOf(direction)) / 100;
            MercatorPoint point = getPoint(location[0], location[1]);
            MercatorVehicle vehicle = this.vehicleMap.get(vid);
            Collection<MercatorGrid> grids = null;
            if(vehicle == null){
                if("1".equals(isOnline) && sp > 0) {
                    grids = getGrids(data, location[0], location[1], deltAngle-direc);
                    vehicle = new MercatorVehicle(this.eventBus, vid, grids, point, Double.parseDouble(String.valueOf(speed)),
                            Long.parseLong(String.valueOf(collectTime)));
                    vehicle.setVersionNo(versionNo);
                    this.vehicleMap.put(vid, vehicle);
                }
            }else{
                //在线，且运动中，则进行网格占用更新，否则全部释放
                if("1".equals(isOnline) && (sp > 0 || vehicle.waitGridSize()>0 || vehicle.isActive())) {
                    grids = getGrids(data, location[0], location[1], deltAngle-direc);
                    vehicle.update(grids, point, Double.parseDouble(String.valueOf(speed)),
                            Long.parseLong(String.valueOf(collectTime)));
                    vehicle.setVersionNo(versionNo);
                    if(sp>0) vehicle.setActive(false);
                }else{//非等待中车辆
                    //车辆已下线 或车辆已停车 释放所有网格
                    vehicle.freeAll();
                }
            }
            if(Objects.isNull(vehicle)){
                return;
            }
            listener.invokeForOrder(vehicle);
            if(vehicle.waitGridSize()>0||(sp==0&&!vehicle.isActive()&&!listener.isContainVin(vid))){//解决关联方停止车辆速度为0,未下命令但为停止状态，
                //车辆暂停任务释放已拥有区域，解决死锁和区域资源占用
                vehicle.freeOwnedGrids();
                data.put("closeArea","");
            }
        }catch (Exception e){
            logger.error(String.format("创建或更新 %s 的MercatorVehicle对象失败:" +
                            "speed: %s, direction: %s, collectTime: %s", vid,  speed, direction, collectTime), e);
        }
    }


    private Collection<MercatorGrid> getGrids( Map<String, Object> data, double lng, double lat, double direc){

        double[] startPoint = {lng, lat};

        double[] delt =  adapterAreaSize(startPoint,direc);
        //寻找车辆当前点，沿车头方向10米的点
        double[] pointO = Coordtransform.getEndPointByTrigonometric(direc, startPoint, 1);
        //pointO，沿车头方向20米的点
        double[] pointE = Coordtransform.getEndPointByTrigonometric(direc, pointO, delt[1]);

        //pointO，沿车头方向垂直90°方向5米的点，
        double[] pointA = Coordtransform.getEndPointByTrigonometric(direc + 90, pointO, delt[0]);
        double[] mercatorA = Coordtransform.WGS84toMercator(pointA[0], pointA[1]);

        //pointO，沿车头方向垂直270°方向5米的点，
        double[] pointD = Coordtransform.getEndPointByTrigonometric(direc + 270, pointO, delt[0]);
        double[] mercatorD = Coordtransform.WGS84toMercator(pointD[0], pointD[1]);

        //pointE，沿车头方向垂直90°方向5米的点，
        double[] pointB = Coordtransform.getEndPointByTrigonometric(direc + 90, pointE, delt[0]);
        double[] mercatorB = Coordtransform.WGS84toMercator(pointB[0], pointB[1]);

        //pointE，沿车头方向垂直270°方向5米的点，
        double[] pointC = Coordtransform.getEndPointByTrigonometric(direc + 270, pointE, delt[0]);
        double[] mercatorC = Coordtransform.WGS84toMercator(pointC[0], pointC[1]);
        List<String> closeArea = new ArrayList<>();
        closeArea.add(pointA[0]+","+pointA[1]);
        closeArea.add(pointB[0]+","+pointB[1]);
        closeArea.add(pointC[0]+","+pointC[1]);
        closeArea.add(pointD[0]+","+pointD[1]);
        data.put("closeArea",closeArea);
      /*  //测试用
        startPoint = Coordtransform.WGS84toGCJ02(startPoint[0], startPoint[1]);
        pointA = Coordtransform.WGS84toGCJ02(pointA[0], pointA[1]);
        pointB = Coordtransform.WGS84toGCJ02(pointB[0], pointB[1]);
        pointC = Coordtransform.WGS84toGCJ02(pointC[0], pointC[1]);
        pointD = Coordtransform.WGS84toGCJ02(pointD[0], pointD[1]);
        System.out.println(String.format("%s: [%s,%s],[%s,%s],[%s,%s],[%s,%s],[%s,%s]", vid, startPoint[0], startPoint[1], pointA[0], pointA[1], pointB[0], pointB[1], pointC[0], pointC[1], pointD[0], pointD[1]));
      */

        //根据车头前方的矩形安全区域的四个顶点墨卡托坐标 查找所占用的网格
        return this.mercatorNet.getChildGrids(new MercatorPoint(mercatorA[0], mercatorA[1]),
                new MercatorPoint(mercatorB[0], mercatorB[1]),
                new MercatorPoint(mercatorC[0], mercatorC[1]),
                new MercatorPoint(mercatorD[0], mercatorD[1]));
    }

    private double[] adapterAreaSize(double[] startPoint,double direction) {

        //判断直行的角度范围
        boolean isStraight = (direction > 276 && direction < 285) || (direction > 367 && direction <376 ) || (direction > 96 && direction < 105) || (direction > 187 && direction < 196);

        if(FieldUtil.isCrossArea(new Point(startPoint[0],startPoint[1]))){
            return new double[]{2,5};
        }else if(!isStraight&& FieldUtil.isSpecialArea(new Point(startPoint[0],startPoint[1]))){
            return new double[]{7,32};
        }else if(!isStraight){
            return new double[]{7,22};
        }
        return new double[]{2,16};
    }

    private MercatorPoint getPoint(double longitude, double latitude){
        double[] mercator = Coordtransform.WGS84toMercator(longitude, latitude);
        return new MercatorPoint(mercator[0], mercator[1]);
    }

    public  void clearOutVersionVehicle(Integer versionNo){
        assert versionNo>-1:"版本号不能小于0";
        List<String> outVersionList = new ArrayList<>();
        this.vehicleMap.forEach((k,v) -> {
            if(v.getVersionNo()<versionNo-240){
                outVersionList.add(k);
            }
        });
        for (String vin:outVersionList){
            MercatorVehicle remove = vehicleMap.remove(vin);
            remove.freeAll();
        }
    }

    @Subscribe
    public void listener(GridFreeEvent event){
        String vid = event.getGrid().getCurrentUser();
        MercatorVehicle vehicle = this.vehicleMap.get(vid);
        if(vehicle != null){
            vehicle.freeWaitGrid(event.getVid(), event.getGrid());
        }else{
            event.getGrid().freeToNext(vid);
        }
    }

}
