package com.ssi.websocket;

import com.alibaba.fastjson.JSON;
import com.ssi.entity.FaultRecord;
import org.apache.commons.collections4.map.PassiveExpiringMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import javax.websocket.*;
import javax.websocket.server.ServerEndpoint;
import java.io.IOException;
import java.util.Date;
import java.util.concurrent.CopyOnWriteArraySet;

/**
 *
 * 异常推送websocket
 */
@ServerEndpoint("/abnormalWebSocket")
@Component
public class AbnormalWebSocketServer {

    private final static Logger log = LoggerFactory.getLogger(AbnormalWebSocketServer.class);
    //静态变量，用来记录当前在线连接数。应该把它设计成线程安全的。
    private static int onlineCount = 0;
    //concurrent包的线程安全Set，用来存放每个客户端对应的MyWebSocket对象。
    private static CopyOnWriteArraySet<AbnormalWebSocketServer> webSocketSet = new CopyOnWriteArraySet<AbnormalWebSocketServer>();

    //与某个客户端的连接会话，需要通过它来给客户端发送数据
    private Session session;

    //忽略消息存储map
    private PassiveExpiringMap<String,Date> ignoreMap = new PassiveExpiringMap<>(180*1000);

    /**
     * 连接建立成功调用的方法
     */
    @OnOpen
    public void onOpen(Session session) {
        this.session = session;
        webSocketSet.add(this);     //加入set中
        addOnlineCount();           //在线数加1
        log.info("有新窗口开始监听:" + session.getId() + ",当前在线人数为" + getOnlineCount());
        try {
            sendMessage("连接成功");
        } catch (IOException e) {
            log.error("websocket IO异常");
        }
    }

    /**
     * 连接关闭调用的方法
     */
    @OnClose
    public void onClose() {
        webSocketSet.remove(this);  //从set中删除
        subOnlineCount();           //在线数减1
        log.info("有一连接关闭！当前在线人数为" + getOnlineCount());
    }

    /**
     * 收到客户端消息后调用的方法
     *
     * @param message 客户端发送过来的消息
     */
    @OnMessage
    public void onMessage(String message, Session session) {
        log.info("收到来自窗口" + session.getId() + "的信息:" + message);
        FaultRecord faultRecord = JSON.parseObject(message, FaultRecord.class);
        addIgnoreMessage(faultRecord);
    }

    public void addIgnoreMessage(FaultRecord faultRecord){
        String vin = faultRecord.getVin();
        String fmi = faultRecord.getFmi();
        String spn = faultRecord.getSpn();
        String key = vin + ":" + fmi + ":" + spn;
        ignoreMap.put(key,new Date());
    }

    public boolean ignoreMessage(FaultRecord faultRecord){
        String vin = faultRecord.getVin();
        String fmi = faultRecord.getFmi();
        String spn = faultRecord.getSpn();
        String key = vin + ":" + fmi + ":" + spn;
        if(!ignoreMap.isEmpty()){
            Date date = ignoreMap.get(key);
            if(date != null){
                ignoreMap.put(key,date);
                return true;
            }
        }
        return false;
    }
    /**
     * @param session
     * @param error
     */
    @OnError
    public void onError(Session session, Throwable error) {
        log.error("发生错误",error);
    }

    /**
     * 实现服务器主动推送
     */
    public void sendMessage(String message) throws IOException {
        this.session.getBasicRemote().sendText(message);
    }

    public static synchronized int getOnlineCount() {
        return onlineCount;
    }

    public static synchronized void addOnlineCount() {
        AbnormalWebSocketServer.onlineCount++;
    }

    public static synchronized void subOnlineCount() {
        AbnormalWebSocketServer.onlineCount--;
    }

    public static void sendInfo(String message) {
        log.info("异常推送websocket收到信息:" + message);
        //群发消息
        for (AbnormalWebSocketServer item : webSocketSet) {
            try {
                item.sendMessage(message);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    public static void sendFaultRecord(FaultRecord faultRecord) {
        //群发消息
        for (AbnormalWebSocketServer item : webSocketSet) {
            try {
                if(!item.ignoreMessage(faultRecord)){
                    item.sendMessage(JSON.toJSONString(faultRecord));
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
}
