package com.ssi.websocket;

import com.alibaba.fastjson.JSON;
import java.io.IOException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;
import javax.websocket.OnClose;
import javax.websocket.OnError;
import javax.websocket.OnMessage;
import javax.websocket.OnOpen;
import javax.websocket.Session;
import javax.websocket.server.PathParam;
import javax.websocket.server.ServerEndpoint;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.springframework.stereotype.Component;

/**
 * 定位websocket服务类,针对单个设备/或车辆
 * <p>type=1 港机定位</p>
 *
 * @author 成东
 * @since 2022-02-23 9:58
 */
@Component
@ServerEndpoint(value = "/locationWebSocket/crane")
@Slf4j
public class LocationWebSocketService {

  private static AtomicInteger onlineCount = new AtomicInteger(0);

  private static ConcurrentMap<String, WebSocketSession> webSocketMap = new ConcurrentHashMap<>();

  public static boolean isPush() {
    for (WebSocketSession w : webSocketMap.values()) {
      if (w.getSession().isOpen()) {
        return true;
      }
    }
    return false;
  }

  /**
   * 连接建立成功调用的方法
   */
  @OnOpen
  public void onOpen(Session session) {
    try {
      log.info(String.format("港机定位websocket打开新会话%s 成功", session.getId()));
      webSocketMap.put(session.getId(), new WebSocketSession(session));
      onlineCount.incrementAndGet();
      sendMessage(session, "连接成功");
    } catch (Exception e) {
      log.error("websocket 打开异常", e);
    }
  }

  /**
   * 连接关闭调用的方法
   */
  @OnClose
  public void onClose(Session session) {
      log.info(String.format("港机定位websocket会话：%s 关闭，当剩余前会话数：%d", session.getId(),
          onlineCount.decrementAndGet()));
      webSocketMap.remove(session.getId());
  }

  /**
   * 收到客户端消息后调用的方法
   *
   * @param message 客户端发送过来的消息
   */
  @OnMessage
  public void onMessage(String message, Session session) {
    log.info(String.format("收到来自会话%s的信息: %s", session.getId(), message));
  }

  /**
   *
   */
  @OnError
  public void onError(Session session, Throwable error) {
    log.error(String.format("会话：%s 发生错误.", session.getId()), error);
  }

  /**
   * 消息广播
   */
  public static void messageBroadcast(Object messages) {
    webSocketMap.values().parallelStream().forEach(webSocketSession -> {
      webSocketSession.sendInfo(messages);
    });

  }

  /**
   * 实现服务器主动推送
   */
  public void sendMessage(Session session, String message) throws IOException {
    session.getBasicRemote().sendText(message);
  }

  private class WebSocketSession {

    @Getter
    private final Session session;


    private final AtomicInteger printSwitch = new AtomicInteger(0);
    private final AtomicInteger successCount = new AtomicInteger(0);
    private final AtomicInteger missCount = new AtomicInteger(0);

    public WebSocketSession(Session session) {
      this.session = session;
    }

    public void sendInfo(Object messages) {
      if (session.isOpen()) {
        try {
          session.getBasicRemote().sendText(JSON.toJSONString(messages));
          successCount.incrementAndGet();
        } catch (IOException e) {
          log.info(String.format("港机定位websocket推送失败: %s, %s ", session.getId(), messages), e);
          missCount.incrementAndGet();
        }
      }
      //定期状态输出
      int minuteOfHour = DateTime.now().getMinuteOfHour();
      int printMinute = 5;
      if (printSwitch.get() < 1 && minuteOfHour % printMinute == 0) {
        log.info(String.format("港机定位webSocket: %s 成功推送次数: %s, 推送失败次数：%s", session.getId(),
            successCount.get(), missCount.get()));
        printSwitch.incrementAndGet();
      } else if (minuteOfHour % printMinute != 0) {
        printSwitch.set(0);
      }
    }
  }

}

