package com.ssi.websocket;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.ssi.entity.VmsVehicle;
import com.ssi.model.VehicleBaseInfoModel;
import com.ssi.service.WebSocketDataBackService;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.websocket.*;
import javax.websocket.server.ServerEndpoint;
import java.io.IOException;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @description
 *   用于脚本数据的回放
 */
@ServerEndpoint(value="/VMSWebBacksocket")
@Component
public class WebBackSocketService {

    private static Logger log = LoggerFactory.getLogger(WebBackSocketService.class);

    private static AtomicInteger onlineCount = new AtomicInteger(0);

    private static Map<String, WebSocketSession> webSocketSet = Maps.newConcurrentMap();
    private Session session;
    private String token;

    private static WebSocketDataBackService webSocketDataBackService;
    private static com.ssi.model.VehicleBaseInfoModel VehicleBaseInfoModel;

    /**
     * 连接建立成功调用的方法
     * */
    @OnOpen
    public void onOpen(Session session) {
        try {
            this.session = session;
            this.token = session.getId();
            WebSocketSession webSocketSession = new WebSocketSession(this);

            WebBackSocketService.webSocketSet.put(token, webSocketSession);
            addOnlineCount();
            log.info(String.format("打开数据回放新会话%s成功。", token));
        }catch (Exception e){
            log.error("打开数据回放websocket 打开异常",  e);
        }
    }

    /**
     * 连接关闭调用的方法
     */
    @OnClose
    public void onClose() {
        webSocketSet.get(this.token).close();
        webSocketSet.remove(this.token);
        subOnlineCount();
        log.info(String.format("数据回放会话：%s 关闭，当剩余前会话数：%s", token,  getOnlineCount()));
    }

    /**
     * 收到客户端消息后调用的方法
     *
     * @param message 客户端发送过来的消息*/
    @OnMessage
    public void onMessage(String message, Session session) {
        try {
            if(StringUtils.isNotBlank(message)) {
                JSONObject messageJson = JSON.parseObject(message);
                if (messageJson != null) {
                    Set<String> vins;
                    String vinStr = messageJson.getString("vins");
                    if (StringUtils.isNotBlank(vinStr)) {
                        vins = Sets.newHashSet(vinStr.split(","));
                    } else {
                        vins = getVins();
                    }

                    if(!vins.isEmpty()) {

                        WebSocketSession webSocketSession
                                = WebBackSocketService.webSocketSet.get(session.getId());

                        Long startTime = messageJson.getLong("startTime");
                        Long stopTime = messageJson.getLong("stopTime");
                        String backorId = messageJson.getString("id");
                        if(backorId == null){
                            backorId = token;
                        }
                        backorId = String.format("%s,%s,%s,%s,%s", backorId, vinStr, "", startTime, stopTime);

                        int status = webSocketDataBackService.playBack(session.getId(), backorId, "", vins, startTime, stopTime);
                        if(status == 0){
                            webSocketSession.backorIds.add(backorId);
                        }
                    }
                }
            }
        } catch (Exception e) {
            log.info(String.format("数据回放客户端参数解析失败：%s", message), e);
        }
    }


    /**
     *
     * @param session
     * @param error
     */
    @OnError
    public void onError(Session session, Throwable error) {
        log.error(String.format("数据回放会话：%s 发生错误.", session.getId()), error);
        webSocketSet.get(session.getId()).close();
    }

    /**
     * 实现服务器主动推送
     */
    public void sendMessage(String message) throws IOException {
        this.session.getBasicRemote().sendText(message);
    }


    /**
     * 消息广播
     * */
    public static void messageBroadcast(String backorId, Set<String> sessions, String json) {
       messageBroadcast(backorId, sessions, json, false);
    }
    public static void messageBroadcast(String backorId, Set<String> sessions, List<String> recordList) {
       messageBroadcast(backorId, sessions, recordList, false);
    }

    public static void messageBroadcast(String backorId, Set<String> sessions, String json, boolean last) {
        for (String session : sessions) {
            webSocketSet.get(session).sendInfo(backorId, json, last);
        }
    }

    public static void messageBroadcast(String backorId, Set<String> sessions, List<String> recordList, boolean last) {
        for (String session : sessions) {
            webSocketSet.get(session).sendInfo(backorId, recordList, last);
        }
    }

    public static int getOnlineCount() {
        return WebBackSocketService.onlineCount.get();
    }

    public static void addOnlineCount() {
        WebBackSocketService.onlineCount.incrementAndGet();
    }

    public static void subOnlineCount() {
        WebBackSocketService.onlineCount.decrementAndGet();
    }

    private class WebSocketSession{

        private WebBackSocketService webSocketService;
        private Set<String> backorIds = Sets.newConcurrentHashSet();

        public WebSocketSession(WebBackSocketService webSocketService) {
            this.webSocketService = webSocketService;
        }

        public synchronized void sendInfo(String backorId, String json, boolean last){
            try {
                if(json != null) {
                    boolean complete = last;
                    if(json.startsWith("#")){
                        complete = true;
                        json = json.substring(1);
                    }
                    webSocketService.sendMessage(String.format("{\"last\":%s,\"timeStamp\":%s, \"car\":%s}",
                            complete, System.currentTimeMillis(), json));
                }
            } catch (IOException e) {
                log.info(String.format("%s数据回放推送失败: %s, 数据量：%s", backorId, webSocketService.token, json), e);

            }finally {
                if(last && backorIds.contains(backorId)){
                    close(backorId);
                }
            }
        }

        public synchronized void sendInfo(String backorId, List<String> recordList, boolean last){
            List<Map<String, Object>> resultList = Lists.newArrayList();
            try {
                if(!recordList.isEmpty()){
                    recordList.stream().forEach(record -> {
                        String json = record;
                        if(json != null) {
                            boolean complete = last;
                            if(json.startsWith("#")){
                                complete = true;
                                json = json.substring(1);
                            }
                            Map<String, Object> res = Maps.newHashMap();
                            res.put("last", complete);
                            res.put("timeStamp", System.currentTimeMillis());
                            res.put("car", JSONObject.parse(json));
                            resultList.add(res);
                        }
                    });
                    webSocketService.sendMessage(JSONObject.toJSONString(resultList));
                }

            } catch (IOException e) {
                log.info(String.format("%s数据回放推送失败: %s, 数据量：%s", backorId, webSocketService.token, JSONObject.toJSONString(resultList)), e);

            }finally {
                if(last && backorIds.contains(backorId)){
                    close(backorId);
                }
            }
        }

        private void close(String backorId){
            webSocketDataBackService.close(webSocketService.token, backorId);
        }
        private void close(){
            backorIds.forEach(backorId -> {
                webSocketDataBackService.close(webSocketService.token, backorId);
            });
        }
    }

    public static boolean hasWebSocketSession() {
        return !webSocketSet.isEmpty();
    }

    private Set<String> getVins(){
        Set<String> vins = null;
            List<VmsVehicle> allVehicleBaseInfo = VehicleBaseInfoModel.getAllVehicleInfo();
            if (allVehicleBaseInfo != null && !allVehicleBaseInfo.isEmpty()) {
                vins = new HashSet<>();
                for (VmsVehicle vehicle : allVehicleBaseInfo) {
                    vins.add(vehicle.getVin());
                }
            }
        return vins;
    }

    @Autowired
    public  void setWebSocketDataBackKafkaModel(WebSocketDataBackService webSocketDataBackService) {
        WebBackSocketService.webSocketDataBackService = webSocketDataBackService;
    }

    @Autowired
    public  void setCismVehicleBaseInfoModel(VehicleBaseInfoModel VehicleBaseInfoModel) {
        WebBackSocketService.VehicleBaseInfoModel = VehicleBaseInfoModel;
    }
}

