package com.ssi.websocket;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import javax.websocket.*;
import javax.websocket.server.ServerEndpoint;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * 大屏及web使用的websocket，推送车辆任务信息
 */
@ServerEndpoint(value = "/VMSWebsocketForScreenList")
@Component
public class WebSocketServiceForScreenList {

    private static Logger log = LoggerFactory.getLogger(WebSocketServiceForScreenList.class);

    private static AtomicInteger onlineCount = new AtomicInteger(0);

    private static Map<String, WebSocketSession> webSocketSet = Maps.newConcurrentMap();
    private Session session;
    private String token;

    /**
     * 连接建立成功调用的方法
     */
    @OnOpen
    public void onOpen(Session session) {
        try {
            this.session = session;
            WebSocketSession webSocketSession = new WebSocketSession();
            webSocketSession.webSocketService = this;

            this.token = session.getId();
            WebSocketServiceForScreenList.webSocketSet.put(token, webSocketSession);
            addOnlineCount();
            log.info(String.format("打开新会话%s成功。", token));
        } catch (Exception e) {
            log.error("websocket 打开异常", e);
        }
    }

    /**
     * 连接关闭调用的方法
     */
    @OnClose
    public void onClose() {
        webSocketSet.remove(this.token);
        subOnlineCount();
        log.info(String.format("会话：%s 关闭，当剩余前会话数：%s", token, getOnlineCount()));
    }

    /**
     * 收到客户端消息后调用的方法
     *
     * @param message 客户端发送过来的消息
     */
    @OnMessage
    public void onMessage(String message, Session session) {
        try {
            if (StringUtils.isNotBlank(message)) {
                JSONObject messageJson = JSON.parseObject(message);
                if (messageJson != null) {
                    String vinStr = messageJson.getString("vins");
                    Set<String> vins;
                    if (StringUtils.isNotBlank(vinStr)) {
                        vins = Sets.newHashSet(vinStr.split(","));
                    } else {
                        vins = Sets.newHashSet();
                    }
                    WebSocketSession webSocketSession = WebSocketServiceForScreenList.webSocketSet.get(session.getId());
                    webSocketSession.setVins(vins);
                    log.info(String.format("收到来自会话%s的信息: %s, vins: %s", session.getId(), message, vins));
                }
            }
        } catch (Exception e) {
            log.info(String.format("客户端参数解析失败：%s", message), e);
        }
    }


    /**
     * @param session
     * @param error
     */
    @OnError
    public void onError(Session session, Throwable error) {
        log.error(String.format("会话：%s 发生错误.", session.getId()), error);
    }

    /**
     * 实现服务器主动推送
     */
    public void sendMessage(String message) throws IOException {
        this.session.getBasicRemote().sendText(message);
    }


    /**
     * 消息广播
     */
    public static void messageBroadcast(Collection<Map<String, Object>> messages) {

        webSocketSet.values()
                .parallelStream()
                .forEach(webSocketSession -> {
                    webSocketSession.sendInfo(messages);
                });

    }

    /**
     * 消息广播
     */
    public static void messageBroadcast(String message) {
        webSocketSet.values().parallelStream().forEach(webSocketSession -> {
                    webSocketSession.sendInfo(message);
                });
    }

    public static int getOnlineCount() {
        return WebSocketServiceForScreenList.onlineCount.get();
    }

    public static void addOnlineCount() {
        WebSocketServiceForScreenList.onlineCount.incrementAndGet();
    }

    public static void subOnlineCount() {
        WebSocketServiceForScreenList.onlineCount.decrementAndGet();
    }

    private class WebSocketSession {

        private WebSocketServiceForScreenList webSocketService;
        private Set<String> vins = new HashSet<>();

        private Object lock = new Object();

        private int printMinute = 5;
        private AtomicInteger printSwitch = new AtomicInteger(0);
        private AtomicInteger successCount = new AtomicInteger(0);
        private AtomicInteger missCount = new AtomicInteger(0);

        public void sendInfo(Collection<Map<String, Object>> messages) {
            List<Map<String, Object>> resultList = Lists.newArrayList();
            Iterator<Map<String, Object>> iter = messages.iterator() ;
            while(iter.hasNext())
            {
                Map<String, Object> res = Maps.newHashMap();
                res.put("timeStamp", System.currentTimeMillis());
                res.put("car", iter.next());
                resultList.add(res);
            }

            try {
                webSocketService.sendMessage(JSON.toJSONString(resultList));
                successCount.incrementAndGet();
            } catch (IOException e) {
                log.info(String.format("推送失败: %s, %s ", webSocketService.token, messages), e);
                missCount.incrementAndGet();
            }


            //定期状态输出
            int minuteOfHour = DateTime.now().getMinuteOfHour();
            if (printSwitch.get() < 1 && minuteOfHour % printMinute == 0) {
                log.info(String.format("大屏list webSocket: %s 成功推送次数: %s, 推送失败次数：%s",
                        webSocketService.token, successCount.get(), missCount.get()));
                printSwitch.incrementAndGet();
            } else if (minuteOfHour % printMinute != 0) {
                printSwitch.set(0);
            }
        }

        public void setVins(Set<String> vins) {
            synchronized (lock) {
                this.vins = vins;
            }
        }

        public void sendInfo(String message) {
            try {
                webSocketService.sendMessage(message);
            } catch (IOException e) {
                log.error(String.format("推送文本消息失败: %s, %s ", webSocketService.token, message), e);
            }
        }
    }

    public static boolean hasWebSocketSession() {
        return !webSocketSet.isEmpty();
    }

}

